import abc
import logging
import math
import keras
import numpy as np
from typing import List, Tuple

from QuizGenerator.premade_questions.cst463.models.matrices import MatrixQuestion
from QuizGenerator.question import Question, QuestionRegistry
from QuizGenerator.contentast import ContentAST, AnswerTypes
from QuizGenerator.constants import MathRanges
from QuizGenerator.mixins import TableQuestionMixin

log = logging.getLogger(__name__)


@QuestionRegistry.register("cst463.word2vec.skipgram")
class word2vec__skipgram(MatrixQuestion, TableQuestionMixin):
  
  @staticmethod
  def skipgram_predict(center_emb, context_embs):
    """
    center_emb: (embed_dim,) - center word embedding
    context_embs: (num_contexts, embed_dim) - context candidate embeddings

    Returns: probabilities (num_contexts,)
    """
    # Compute dot products (logits)
    logits = context_embs @ center_emb
    
    # Softmax
    exp_logits = np.exp(logits)
    probs = exp_logits / exp_logits.sum()
    
    return logits, probs
  
  def refresh(self, *args, **kwargs):
    super().refresh(*args, **kwargs)
    self.rng = np.random.RandomState(kwargs.get("rng_seed", None))
    
    embed_dim = kwargs.get("embed_dim", 3)
    num_contexts = kwargs.get("num_contexts", 3)
    
    # Vocabulary pool
    vocab = ['cat', 'dog', 'run', 'jump', 'happy', 'sad', 'tree', 'house',
             'walk', 'sleep', 'fast', 'slow', 'big', 'small']
    
    # Sample words
    self.selected_words = self.rng.choice(vocab, size=num_contexts + 1, replace=False)
    self.center_word = self.selected_words[0]
    self.context_words = self.selected_words[1:]
    
    # Small integer embeddings

    self.center_emb = self.get_rounded_matrix((embed_dim,), -2, 3)
    self.context_embs = self.get_rounded_matrix((num_contexts, embed_dim), -2, 3)
    
    self.logits, self.probs = self.skipgram_predict(self.center_emb, self.context_embs)

    ## Answers:
    # center_word, center_emb, context_words, context_embs, logits, probs
    self.answers["logits"] = AnswerTypes.Vector(self.logits, label="Logits")
    most_likely_idx = np.argmax(self.probs)
    most_likely_word = self.context_words[most_likely_idx]
    self.answers["center_word"] = AnswerTypes.String(most_likely_word, label="Most likely context word")
    
    
    return True
  
  def _get_body(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question body and collect answers."""
    body = ContentAST.Section()
    answers = []

    body.add_element(
      ContentAST.Paragraph([
        f"Given center word: `{self.center_word}` with embedding {self.center_emb}, compute the skip-gram probabilities for each context word and identify the most likely one."
      ])
    )
    body.add_elements([
      ContentAST.Paragraph([ContentAST.Text(f"`{w}` : "), str(e)]) for w, e in zip(self.context_words, self.context_embs)
    ])

    answers.append(self.answers["logits"])
    answers.append(self.answers["center_word"])
    body.add_elements([
      ContentAST.LineBreak(),
      self.answers["logits"],
      ContentAST.LineBreak(),
      self.answers["center_word"]
    ])

    log.debug(f"output: {self.logits}")
    log.debug(f"weights: {self.probs}")

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body
  
  def _get_explanation(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question explanation."""
    explanation = ContentAST.Section()
    digits = ContentAST.Answer.DEFAULT_ROUNDING_DIGITS

    explanation.add_element(
      ContentAST.Paragraph([
        "In the skip-gram model, we predict context words given a center word by computing dot products between embeddings and applying softmax."
      ])
    )

    # Step 1: Show embeddings
    explanation.add_element(
      ContentAST.Paragraph([
        ContentAST.Text("Step 1: Given embeddings", emphasis=True)
      ])
    )

    # Format center embedding
    center_emb_str = "[" + ", ".join([f"{x:.{digits}f}" for x in self.center_emb]) + "]"
    explanation.add_element(
      ContentAST.Paragraph([
        f"Center word `{self.center_word}`: {center_emb_str}"
      ])
    )

    explanation.add_element(
      ContentAST.Paragraph([
        "Context words:"
      ])
    )

    for i, (word, emb) in enumerate(zip(self.context_words, self.context_embs)):
      emb_str = "[" + ", ".join([f"{x:.2f}" for x in emb]) + "]"
      explanation.add_element(
        ContentAST.Paragraph([
          f"`{word}`: {emb_str}"
        ])
      )

    # Step 2: Compute logits (dot products)
    explanation.add_element(
      ContentAST.Paragraph([
        ContentAST.Text("Step 2: Compute logits (dot products)", emphasis=True)
      ])
    )

    # Show ONE example
    explanation.add_element(
      ContentAST.Paragraph([
        f"Example: Logit for `{self.context_words[0]}`"
      ])
    )

    context_emb = self.context_embs[0]
    dot_product_terms = " + ".join([f"({self.center_emb[j]:.2f} \\times {context_emb[j]:.2f})"
                                    for j in range(len(self.center_emb))])
    logit_val = self.logits[0]

    explanation.add_element(
      ContentAST.Equation(f"{dot_product_terms} = {logit_val:.2f}")
    )

    logits_str = "[" + ", ".join([f"{x:.2f}" for x in self.logits]) + "]"
    explanation.add_element(
      ContentAST.Paragraph([
        f"All logits: {logits_str}"
      ])
    )

    # Step 3: Apply softmax
    explanation.add_element(
      ContentAST.Paragraph([
        ContentAST.Text("Step 3: Apply softmax to get probabilities", emphasis=True)
      ])
    )

    exp_logits = np.exp(self.logits)
    sum_exp = exp_logits.sum()

    exp_terms = " + ".join([f"e^{{{l:.{digits}f}}}" for l in self.logits])

    explanation.add_element(
      ContentAST.Equation(f"\\text{{denominator}} = {exp_terms} = {sum_exp:.{digits}f}")
    )

    explanation.add_element(
      ContentAST.Paragraph([
        "Probabilities:"
      ])
    )

    for i, (word, prob) in enumerate(zip(self.context_words, self.probs)):
      explanation.add_element(
        ContentAST.Equation(f"P(\\text{{{word}}}) = \\frac{{e^{{{self.logits[i]:.{digits}f}}}}}{{{sum_exp:.{digits}f}}} = {prob:.{digits}f}")
      )

    # Step 4: Identify most likely
    most_likely_idx = np.argmax(self.probs)
    most_likely_word = self.context_words[most_likely_idx]

    explanation.add_element(
      ContentAST.Paragraph([
        ContentAST.Text("Conclusion:", emphasis=True),
        f" The most likely context word is `{most_likely_word}` with probability {self.probs[most_likely_idx]:.{digits}f}"
      ])
    )

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation




#!env python
from __future__ import annotations

import abc
import difflib
import logging

from QuizGenerator.question import Question, QuestionRegistry
from QuizGenerator.contentast import ContentAST, AnswerTypes
from QuizGenerator.mixins import TableQuestionMixin, BodyTemplatesMixin

log = logging.getLogger(__name__)


class IOQuestion(Question, abc.ABC):
  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.IO)
    super().__init__(*args, **kwargs)
  

@QuestionRegistry.register()
class HardDriveAccessTime(IOQuestion, TableQuestionMixin, BodyTemplatesMixin):
  
  def refresh(self, *args, **kwargs):
    super().refresh(*args, **kwargs)
    
    self.hard_drive_rotation_speed = 100 * self.rng.randint(36, 150)  # e.g. 3600rpm to 15000rpm
    self.seek_delay = float(round(self.rng.randrange(3, 20), 2))
    self.transfer_rate = self.rng.randint(50, 300)
    self.number_of_reads = self.rng.randint(1, 20)
    self.size_of_reads = self.rng.randint(1, 10)
    
    self.rotational_delay = (1 / self.hard_drive_rotation_speed) * (60 / 1) * (1000 / 1) * (1/2)
    self.access_delay = self.rotational_delay + self.seek_delay
    self.transfer_delay = 1000 * (self.size_of_reads * self.number_of_reads) / 1024 / self.transfer_rate
    self.disk_access_delay = self.access_delay * self.number_of_reads + self.transfer_delay
    
    self.answers.update({
      "answer__rotational_delay"  : AnswerTypes.Float(self.rotational_delay),
      "answer__access_delay"      : AnswerTypes.Float(self.access_delay),
      "answer__transfer_delay"    : AnswerTypes.Float(self.transfer_delay),
      "answer__disk_access_delay" : AnswerTypes.Float(self.disk_access_delay),
    })
  
  def _get_body(self, *args, **kwargs):
    """Build question body and collect answers."""
    answers = [
      self.answers["answer__rotational_delay"],
      self.answers["answer__access_delay"],
      self.answers["answer__transfer_delay"],
      self.answers["answer__disk_access_delay"],
    ]

    # Create parameter info table using mixin
    parameter_info = {
      "Hard Drive Rotation Speed": f"{self.hard_drive_rotation_speed}RPM",
      "Seek Delay": f"{self.seek_delay}ms",
      "Transfer Rate": f"{self.transfer_rate}MB/s",
      "Number of Reads": f"{self.number_of_reads}",
      "Size of Reads": f"{self.size_of_reads}KB"
    }

    parameter_table = self.create_info_table(parameter_info)

    # Create answer table with multiple rows using mixin
    answer_rows = [
      {"Variable": "Rotational Delay", "Value": "answer__rotational_delay"},
      {"Variable": "Access Delay", "Value": "answer__access_delay"},
      {"Variable": "Transfer Delay", "Value": "answer__transfer_delay"},
      {"Variable": "Total Disk Access Delay", "Value": "answer__disk_access_delay"}
    ]

    answer_table = self.create_answer_table(
      headers=["Variable", "Value"],
      data_rows=answer_rows,
      answer_columns=["Value"]
    )

    # Use mixin to create complete body with both tables
    intro_text = "Given the information below, please calculate the following values."

    instructions = (
      f"Make sure that if you round your answers you use the unrounded values for your final calculations, "
      f"otherwise you may introduce error into your calculations."
      f"(i.e. don't use your rounded answers to calculate your overall answer)"
    )

    body = self.create_parameter_calculation_body(
      intro_text=intro_text,
      parameter_table=parameter_table,
      answer_table=answer_table,
      additional_instructions=instructions
    )

    return body, answers

  def get_body(self, *args, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(*args, **kwargs)
    return body

  def _get_explanation(self):
    explanation = ContentAST.Section()
    
    explanation.add_element(
      ContentAST.Paragraph([
        "To calculate the total disk access time (or \"delay\"), "
        "we should first calculate each of the individual parts.",
        r"Since we know that  $t_{total} = (\text{# of reads}) \cdot t_{access} + t_{transfer}$"
        r"we therefore need to calculate $t_{access}$ and  $t_{transfer}$, where "
        r"$t_{access} = t_{rotation} + t_{seek}$.",
      ])
    )
    
    explanation.add_elements([
      ContentAST.Paragraph(["Starting with the rotation delay, we calculate:"]),
      ContentAST.Equation(
        "t_{rotation} = "
        + f"\\frac{{1 minute}}{{{self.hard_drive_rotation_speed}revolutions}}"
        + r"\cdot \frac{60 seconds}{1 minute} \cdot \frac{1000 ms}{1 second} \cdot \frac{1 revolution}{2} = "
        + f"{self.rotational_delay:0.2f}ms",
      )
    ])
    
    explanation.add_elements([
      ContentAST.Paragraph([
        "Now we can calculate:",
      ]),
      ContentAST.Equation(
        f"t_{{access}} "
        f"= t_{{rotation}} + t_{{seek}} "
        f"= {self.rotational_delay:0.2f}ms + {self.seek_delay:0.2f}ms = {self.access_delay:0.2f}ms"
      )
    ])
    
    explanation.add_elements([
      ContentAST.Paragraph([r"Next we need to calculate our transfer delay, $t_{transfer}$, which we do as:"]),
      ContentAST.Equation(
        f"t_{{transfer}} "
        f"= \\frac{{{self.number_of_reads} \\cdot {self.size_of_reads}KB}}{{1}} \\cdot \\frac{{1MB}}{{1024KB}} "
        f"\\cdot \\frac{{1 second}}{{{self.transfer_rate}MB}} \\cdot \\frac{{1000ms}}{{1second}} "
        f"= {self.transfer_delay:0.2}ms"
      )
    ])
    
    explanation.add_elements([
      ContentAST.Paragraph(["Putting these together we get:"]),
      ContentAST.Equation(
        f"t_{{total}} "
        f"= \\text{{(# reads)}} \\cdot t_{{access}} + t_{{transfer}} "
        f"= {self.number_of_reads} \\cdot {self.access_delay:0.2f} + {self.transfer_delay:0.2f} "
        f"= {self.disk_access_delay:0.2f}ms")
    ])
    return explanation, []

  def get_explanation(self) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation()
    return explanation


@QuestionRegistry.register()
class INodeAccesses(IOQuestion, TableQuestionMixin, BodyTemplatesMixin):
  
  def refresh(self, *args, **kwargs):
    super().refresh(*args, **kwargs)
    
    # Calculating this first to use blocksize as an even multiple of it
    self.inode_size = 2**self.rng.randint(6, 10)
    
    self.block_size = self.inode_size * self.rng.randint(8, 20)
    self.inode_number = self.rng.randint(0, 256)
    self.inode_start_location = self.block_size * self.rng.randint(2, 5)
    
    self.inode_address = self.inode_start_location + self.inode_number * self.inode_size
    self.inode_block = self.inode_address // self.block_size
    self.inode_address_in_block = self.inode_address % self.block_size
    self.inode_index_in_block = int(self.inode_address_in_block / self.inode_size)
    
    self.answers.update({
      "answer__inode_address": AnswerTypes.Int(self.inode_address),
      "answer__inode_block": AnswerTypes.Int(self.inode_block),
      "answer__inode_address_in_block": AnswerTypes.Int(self.inode_address_in_block),
      "answer__inode_index_in_block": AnswerTypes.Int(self.inode_index_in_block),
    })
  
  def _get_body(self):
    """Build question body and collect answers."""
    answers = [
      self.answers["answer__inode_address"],
      self.answers["answer__inode_block"],
      self.answers["answer__inode_address_in_block"],
      self.answers["answer__inode_index_in_block"],
    ]

    # Create parameter info table using mixin
    parameter_info = {
      "Block Size": f"{self.block_size} Bytes",
      "Inode Number": f"{self.inode_number}",
      "Inode Start Location": f"{self.inode_start_location} Bytes",
      "Inode size": f"{self.inode_size} Bytes"
    }

    parameter_table = self.create_info_table(parameter_info)

    # Create answer table with multiple rows using mixin
    answer_rows = [
      {"Variable": "Inode address", "Value": "answer__inode_address"},
      {"Variable": "Block containing inode", "Value": "answer__inode_block"},
      {"Variable": "Inode address (offset) within block", "Value": "answer__inode_address_in_block"},
      {"Variable": "Inode index within block", "Value": "answer__inode_index_in_block"}
    ]

    answer_table = self.create_answer_table(
      headers=["Variable", "Value"],
      data_rows=answer_rows,
      answer_columns=["Value"]
    )

    # Use mixin to create complete body with both tables
    intro_text = "Given the information below, please calculate the following values."

    body = self.create_parameter_calculation_body(
      intro_text=intro_text,
      parameter_table=parameter_table,
      answer_table=answer_table,
      # additional_instructions=instructions
    )

    return body, answers

  def get_body(self) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body()
    return body

  def _get_explanation(self):
    explanation = ContentAST.Section()
    
    explanation.add_element(
      ContentAST.Paragraph([
        "If we are given an inode number, there are a few steps that we need to take to load the actual inode.  "
        "These consist of determining the address of the inode, which block would contain it, "
        "and then its address within the block.",
        "To find the inode address, we calculate:",
      ])
    )
    
    explanation.add_element(
      ContentAST.Equation.make_block_equation__multiline_equals(
        r"(\text{Inode address})",
        [
          r"(\text{Inode Start Location}) + (\text{inode #}) \cdot (\text{inode size})",
          f"{self.inode_start_location} + {self.inode_number} \\cdot {self.inode_size}",
          f"{self.inode_address}"
        ])
    )
    
    explanation.add_element(
      ContentAST.Paragraph([
        "Next, we us this to figure out what block the inode is in.  "
        "We do this directly so we know what block to load, "
        "thus minimizing the number of loads we have to make.",
      ])
    )
    explanation.add_element(ContentAST.Equation.make_block_equation__multiline_equals(
      r"\text{Block containing inode}",
      [
        r"(\text{Inode address}) \mathbin{//} (\text{block size})",
        f"{self.inode_address} \\mathbin{{//}} {self.block_size}",
        f"{self.inode_block}"
      ]
    ))
    
    explanation.add_element(
      ContentAST.Paragraph([
        "When we load this block, we now have in our system memory "
        "(remember, blocks on the hard drive are effectively useless to us until they're in main memory!), "
        "the inode, so next we need to figure out where it is within that block."
        "This means that we'll need to find the offset into this block.  "
        "We'll calculate this both as the offset in bytes, and also in number of inodes, "
        "since we can use array indexing.",
      ])
    )
    
    explanation.add_element(ContentAST.Equation.make_block_equation__multiline_equals(
      r"\text{offset within block}",
      [
        r"(\text{Inode address}) \bmod (\text{block size})",
        f"{self.inode_address} \\bmod {self.block_size}",
        f"{self.inode_address_in_block}"
      ]
    ))
    
    explanation.add_element(
      ContentAST.Text("Remember that `mod` is the same as `%`, the modulo operation.")
    )
    
    explanation.add_element(ContentAST.Paragraph(["and"]))
      
    explanation.add_element(ContentAST.Equation.make_block_equation__multiline_equals(
      r"\text{index within block}",
      [
        r"\dfrac{\text{offset within block}}{\text{inode size}}",
        f"\\dfrac{{{self.inode_address_in_block}}}{{{self.inode_size}}}",
        f"{self.inode_index_in_block}"
      ]
    ))

    return explanation, []

  def get_explanation(self) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation()
    return explanation


@QuestionRegistry.register()
class VSFS_states(IOQuestion):

  from .ostep13_vsfs import fs as vsfs
  
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.answer_kind = ContentAST.Answer.CanvasAnswerKind.MULTIPLE_DROPDOWN
    
    self.num_steps = kwargs.get("num_steps", 10)
  
  def refresh(self, *args, **kwargs):
    super().refresh(*args, **kwargs)
    
    fs = self.vsfs(4, 4, self.rng)
    operations = fs.run_for_steps(self.num_steps)
    
    self.start_state = operations[-1]["start_state"]
    self.end_state = operations[-1]["end_state"]
    
    wrong_answers = list(filter(
      lambda o: o != operations[-1]["cmd"],
      map(
        lambda o: o["cmd"],
        operations
      )
    ))
    self.rng.shuffle(wrong_answers)
    
    self.answers["answer__cmd"] = ContentAST.Answer.dropdown(
      f"{operations[-1]['cmd']}",
      baffles=list(set([op['cmd'] for op in operations[:-1] if op != operations[-1]['cmd']])),
      label="Command"
    )
  
  def _get_body(self):
    """Build question body and collect answers."""
    answers = [self.answers["answer__cmd"]]

    body = ContentAST.Section()

    body.add_element(ContentAST.Paragraph(["What operation happens between these two states?"]))

    body.add_element(
      ContentAST.Code(
        self.start_state,
        make_small=True
      )
    )

    body.add_element(ContentAST.AnswerBlock(self.answers["answer__cmd"]))

    body.add_element(
      ContentAST.Code(
        self.end_state,
        make_small=True
      )
    )

    return body, answers

  def get_body(self) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body()
    return body

  def _get_explanation(self):
    explanation = ContentAST.Section()
    
    log.debug(f"self.start_state: {self.start_state}")
    log.debug(f"self.end_state: {self.end_state}")
    
    explanation.add_elements([
      ContentAST.Paragraph([
        "The key thing to pay attention to when solving these problems is where there are differences between the start state and the end state.",
        "In this particular problem, we can see that these lines are different:"
      ])
    ])
    
    chunk_to_add = []
    lines_that_changed = []
    for start_line, end_line in zip(self.start_state.split('\n'), self.end_state.split('\n')):
      if start_line == end_line:
        continue
      lines_that_changed.append((start_line, end_line))
      chunk_to_add.append(
        f" - `{start_line}` -> `{end_line}`"
      )
    
    explanation.add_element(
      ContentAST.Paragraph(chunk_to_add)
    )
    
    chunk_to_add = [
      "A great place to start is to check to see if the bitmaps have changed as this can quickly tell us a lot of information"
    ]
    
    inode_bitmap_lines = list(filter(lambda s: "inode bitmap" in s[0], lines_that_changed))
    data_bitmap_lines = list(filter(lambda s: "data bitmap" in s[0], lines_that_changed))
    
    def get_bitmap(line: str) -> str:
      log.debug(f"line: {line}")
      return line.split()[-1]
    
    def highlight_changes(a: str, b: str) -> str:
      matcher = difflib.SequenceMatcher(None, a, b)
      result = []
      
      for tag, i1, i2, j1, j2 in matcher.get_opcodes():
        if tag == "equal":
          result.append(b[j1:j2])
        elif tag in ("insert", "replace"):
          result.append(f"***{b[j1:j2]}***")
        # for "delete", do nothing since text is removed
      
      return "".join(result)
    
    if len(inode_bitmap_lines) > 0:
      inode_bitmap_lines = inode_bitmap_lines[0]
      chunk_to_add.append(f"The inode bitmap lines have changed from {get_bitmap(inode_bitmap_lines[0])} to {get_bitmap(inode_bitmap_lines[1])}.")
      if get_bitmap(inode_bitmap_lines[0]).count('1') < get_bitmap(inode_bitmap_lines[1]).count('1'):
        chunk_to_add.append("We can see that we have added an inode, so we have either called `creat` or `mkdir`.")
      else:
        chunk_to_add.append("We can see that we have removed an inode, so we have called `unlink`.")
    
    if len(data_bitmap_lines) > 0:
      data_bitmap_lines = data_bitmap_lines[0]
      chunk_to_add.append(f"The inode bitmap lines have changed from {get_bitmap(data_bitmap_lines[0])} to {get_bitmap(data_bitmap_lines[1])}.")
      if get_bitmap(data_bitmap_lines[0]).count('1') < get_bitmap(data_bitmap_lines[1]).count('1'):
        chunk_to_add.append("We can see that we have added a data block, so we have either called `mkdir` or `write`.")
      else:
        chunk_to_add.append("We can see that we have removed a data block, so we have `unlink`ed a file.")
    
    if len(data_bitmap_lines) == 0 and len(inode_bitmap_lines) == 0:
      chunk_to_add.append("If they have not changed, then we know we must have eithered called `link` or `unlink` and must check the references.")
      
    explanation.add_element(
      ContentAST.Paragraph(chunk_to_add)
    )
    
    explanation.add_elements([
      ContentAST.Paragraph(["The overall changes are highlighted with `*` symbols below"])
    ])
    
    explanation.add_element(
      ContentAST.Code(
        highlight_changes(self.start_state, self.end_state)
      )
    )

    return explanation, []

  def get_explanation(self) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation()
    return explanation
  


"""
Canvas LMS integration for QuizGenerator.

Vendored from LMSInterface v0.1.0 (2025-11-24)

This module provides Canvas API integration for uploading quizzes
and managing course content.
"""

__version__ = "0.1.0"
__vendored_from__ = "LMSInterface"
__vendored_date__ = "2025-11-24"




#!env python
from __future__ import annotations

import abc
import enum
import itertools
from typing import List, Dict, Optional, Tuple, Any

from QuizGenerator.question import QuestionRegistry, Question

from QuizGenerator.contentast import ContentAST, AnswerTypes

import logging
log = logging.getLogger(__name__)


class LanguageQuestion(Question, abc.ABC):
  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.LANGUAGES)
    super().__init__(*args, **kwargs)

class BNF:
  
  class Grammar:
    def __init__(self, symbols, start_symbol=None):
      self.start_symbol = start_symbol if start_symbol is not None else symbols[0]
      self.symbols = symbols
  
    def generate(self, include_spaces=False, early_exit=False, early_exit_min_iterations=5):
      curr_symbols : List[BNF.Symbol] = [self.start_symbol]
      prev_symbols: List[BNF.Symbol] = curr_symbols
      
      iteration_count = 0
      # Check to see if we have any non-terminals left
      while any(map(lambda s: s.kind == BNF.Symbol.Kind.NonTerminal, curr_symbols)):
        # Grab the previous symbols in case we are early exitting
        prev_symbols = curr_symbols
        
        # Walk through the current symbols and build a new list of symbols from it
        next_symbols : List[BNF.Symbol] = []
        for symbol in curr_symbols:
          next_symbols.extend(symbol.expand())
        curr_symbols = next_symbols
        
        iteration_count += 1
        
        if early_exit and iteration_count > early_exit_min_iterations:
          break
        
      if early_exit:
        # If we are doing an early exit then we are going to return things with non-terminals
        curr_symbols = prev_symbols
      
      # Take all the current symbols and combine them
      return ('' if not include_spaces else ' ').join([str(s) for s in curr_symbols])
    
    def print(self):
      for symbol in self.symbols:
        print(symbol.get_full_str())
        
    def get_grammar_string(self):
      lines = []
      for symbol in self.symbols:
        lines.append(f"{symbol.get_full_str()}")

      return '\n'.join(lines)
      
  class Symbol:
    
    class Kind(enum.Enum):
      NonTerminal = enum.auto()
      Terminal = enum.auto()
      
    def __init__(self, symbol : str, kind : Kind, rng):
      self.symbol = symbol
      self.kind = kind
      self.productions : List[BNF.Production] = [] # productions
      self.rng = rng
    
    def __str__(self):
      # if self.kind == BNF.Symbol.Kind.NonTerminal:
      #   return f"`{self.symbol}`"
      # else:
      #   return f"{self.symbol}"
      return f"{self.symbol}"
    
    def get_full_str(self):
      return f"{self.symbol} ::= {' | '.join([str(p) for p in self.productions])}"
    
    def add_production(self, production: BNF.Production):
      self.productions.append(production)
    
    def expand(self) -> List[BNF.Symbol]:
      if self.kind == BNF.Symbol.Kind.Terminal:
        return [self]
      return self.rng.choice(self.productions).production
  
  class Production:
    def __init__(self, production_line, nonterminal_symbols: Dict[str, BNF.Symbol], rng):
      if len(production_line.strip()) == 0:
        self.production = []
      else:
        self.production = [
          (nonterminal_symbols.get(symbol, BNF.Symbol(symbol, BNF.Symbol.Kind.Terminal, rng=rng)))
          for symbol in production_line.split(' ')
        ]
      
    def __str__(self):
      if len(self.production) == 0:
        return '""'
      return f"{' '.join([str(s) for s in self.production])}"
  
  
  @staticmethod
  def parse_bnf(grammar_str, rng) -> BNF.Grammar:
    
    # Figure out all the nonterminals and create a Token for them
    terminal_symbols = {}
    start_symbol = None
    for line in grammar_str.strip().splitlines():
      if "::=" in line:
        non_terminal_str, _ = line.split("::=", 1)
        non_terminal_str = non_terminal_str.strip()
        
        terminal_symbols[non_terminal_str] = BNF.Symbol(non_terminal_str, BNF.Symbol.Kind.NonTerminal, rng=rng)
        if start_symbol is None:
          start_symbol = terminal_symbols[non_terminal_str]
    
    # Parse the grammar statement
    for line in grammar_str.strip().splitlines():
      if "::=" in line:
        # Split the line into non-terminal and its expansions
        non_terminal_str, expansions = line.split("::=", 1)
        non_terminal_str = non_terminal_str.strip()
        
        non_terminal = terminal_symbols[non_terminal_str]
        
        for production_str in expansions.split('|'):
          production_str = production_str.strip()
          non_terminal.add_production(BNF.Production(production_str, terminal_symbols, rng=rng))
    bnf_grammar = BNF.Grammar(list(terminal_symbols.values()), start_symbol)
    return bnf_grammar


@QuestionRegistry.register("LanguageQuestion")
class ValidStringsInLanguageQuestion(LanguageQuestion):
  MAX_TRIES = 1000
  
  def __init__(self, grammar_str_good: Optional[str] = None, grammar_str_bad: Optional[str] = None, *args, **kwargs):
    # Preserve question-specific params for QR code config
    if grammar_str_good is not None:
      kwargs['grammar_str_good'] = grammar_str_good
    if grammar_str_bad is not None:
      kwargs['grammar_str_bad'] = grammar_str_bad

    super().__init__(*args, **kwargs)

    # Store whether grammars are fixed (provided) or should be randomized
    self.fixed_grammars = grammar_str_good is not None and grammar_str_bad is not None
    if self.fixed_grammars:
      self.grammar_str_good = grammar_str_good
      self.grammar_str_bad = grammar_str_bad
      self.include_spaces = kwargs.get("include_spaces", False)
      self.MAX_LENGTH = kwargs.get("max_length", 30)
      self.grammar_good = BNF.parse_bnf(self.grammar_str_good, self.rng)
      self.grammar_bad = BNF.parse_bnf(self.grammar_str_bad, self.rng)

    self.num_answer_options = kwargs.get("num_answer_options", 4)
    self.num_answer_blanks = kwargs.get("num_answer_blanks", 4)

  def _select_random_grammar(self):
    """Select and set a random grammar. Called from refresh() to ensure each PDF gets different grammar."""
    which_grammar = self.rng.choice(range(4))

    if which_grammar == 0:
      # todo: make a few different kinds of grammars that could be picked
      self.grammar_str_good = """
        <expression> ::= <term> | <expression> + <term> | <expression> - <term>
        <term>       ::= <factor> | <term> * <factor> | <term> / <factor>
        <factor>     ::= <number>
        <number>     ::= <digit> | <number> <digit>
        <digit>      ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
      """
      # Adding in a plus to number
      self.grammar_str_bad = """
        <expression> ::= <term> | <expression> + <term> | <expression> - <term>
        <term>       ::= <factor> | <term> * <factor> | <term> / <factor>
        <factor>     ::= <number>
        <number>     ::= <digit> + | <digit> <number>
        <digit>      ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
      """
      self.include_spaces = False
      self.MAX_LENGTH = 30
    elif which_grammar == 1:
      self.grammar_str_good = """
        <sentence> ::= <subject> <verb> <object>
        <subject> ::= The cat | A dog | The bird | A child | <adjective> <animal>
        <animal> ::= cat | dog | bird | child
        <adjective> ::= happy | sad | angry | playful
        <verb> ::= chases | sees | hates | loves
        <object> ::= the ball | the toy | the tree | <adjective> <object>
      """
      self.grammar_str_bad = """
        <sentence> ::= <subject> <verb> <object>
        <subject> ::= The human | The dog | A bird | Some child | A <adjective> <animal>
        <animal> ::= cat | dog | bird | child
        <adjective> ::= happy | sad | angry | playful
        <verb> ::= chases | sees | hates | loves
        <object> ::= the ball | the toy | the tree | <adjective> <object>
      """
      self.include_spaces = True
      self.MAX_LENGTH = 100
    elif which_grammar == 2:
      self.grammar_str_good = """
        <poem> ::= <line> | <line> <poem>
        <line> ::= <subject> <verb> <object> <modifier>
        <subject> ::= whispers | shadows | dreams | echoes | <compound-subject>
        <compound-subject> ::= <subject> and <subject>
        <verb> ::= dance | dissolve | shimmer | collapse | <compound-verb>
        <compound-verb> ::= <verb> then <verb>
        <object> ::= beneath | between | inside | around | <compound-object>
        <compound-object> ::= <object> through <object>
        <modifier> ::= silently | violently | mysteriously | endlessly | <recursive-modifier>
        <recursive-modifier> ::= <modifier> and <modifier>
      """
      self.grammar_str_bad = """
        <bad-poem> ::= <almost-valid-line> | <bad-poem> <bad-poem>
        <almost-valid-line> ::= <tricky-subject> <tricky-verb> <tricky-object> <tricky-modifier>
        <tricky-subject> ::= whispers | shadows and and | <duplicate-subject>
        <duplicate-subject> ::= whispers whispers
        <tricky-verb> ::= dance | <incorrect-verb-placement> | <verb-verb>
        <incorrect-verb-placement> ::= dance dance
        <verb-verb> ::= dance whispers
        <tricky-object> ::= beneath | <object-verb-swap> | <duplicate-object>
        <object-verb-swap> ::= dance beneath
        <duplicate-object> ::= beneath beneath
        <tricky-modifier> ::= silently | <modifier-subject-swap> | <duplicate-modifier>
        <modifier-subject-swap> ::= whispers silently
        <duplicate-modifier> ::= silently silently
      """
      self.include_spaces = True
      self.MAX_LENGTH = 100
    elif which_grammar == 3:
      self.grammar_str_good = """
        <A> ::= a <B> a |
        <B> ::= b <C> b |
        <C> ::= c <A> c |
      """
      self.grammar_str_bad = """
        <A> ::= a <B> c
        <B> ::= b <C> a |
        <C> ::= c <A> b |
      """
      self.include_spaces = False
      self.MAX_LENGTH = 100

    self.grammar_good = BNF.parse_bnf(self.grammar_str_good, self.rng)
    self.grammar_bad = BNF.parse_bnf(self.grammar_str_bad, self.rng)
  
  def refresh(self, *args, **kwargs):
    super().refresh(*args, **kwargs)

    # Re-select random grammar for each refresh if not using fixed grammars
    if not self.fixed_grammars:
      self._select_random_grammar()

    self.answers = {}

    # Create answers with proper ContentAST.Answer signature
    # value is the generated string, correct indicates if it's a valid answer
    good_string = self.grammar_good.generate(self.include_spaces)
    self.answers["answer_good"] = ContentAST.Answer(
      value=good_string,
      kind=ContentAST.Answer.CanvasAnswerKind.MULTIPLE_ANSWER,
      correct=True
    )

    bad_string = self.grammar_bad.generate(self.include_spaces)
    self.answers["answer_bad"] = ContentAST.Answer(
      value=bad_string,
      kind=ContentAST.Answer.CanvasAnswerKind.MULTIPLE_ANSWER,
      correct=False
    )

    bad_early_string = self.grammar_bad.generate(self.include_spaces, early_exit=True)
    self.answers["answer_bad_early"] = ContentAST.Answer(
      value=bad_early_string,
      kind=ContentAST.Answer.CanvasAnswerKind.MULTIPLE_ANSWER,
      correct=False
    )

    answer_text_set = {a.value for a in self.answers.values()}
    num_tries = 0
    while len(self.answers) < 10 and num_tries < self.MAX_TRIES:

      correct = self.rng.choice([True, False])
      if not correct:
        early_exit = self.rng.choice([True, False])
      else:
        early_exit = False

      generated_string = (
        self.grammar_good
        if correct or early_exit
        else self.grammar_bad
      ).generate(self.include_spaces, early_exit=early_exit)

      is_correct = correct and not early_exit

      if len(generated_string) < self.MAX_LENGTH and generated_string not in answer_text_set:
        self.answers[f"answer_{num_tries}"] = ContentAST.Answer(
          value=generated_string,
          kind=ContentAST.Answer.CanvasAnswerKind.MULTIPLE_ANSWER,
          correct=is_correct
        )
        answer_text_set.add(generated_string)
      num_tries += 1
    
    # Generate answers that will be used only for the latex version.
    self.featured_answers = {
      self.grammar_good.generate(),
      self.grammar_bad.generate(),
      self.grammar_good.generate(early_exit=True)
    }
    while len(self.featured_answers) < self.num_answer_options:
      self.featured_answers.add(
        self.rng.choice([
          lambda: self.grammar_good.generate(),
          lambda: self.grammar_bad.generate(),
          lambda: self.grammar_good.generate(early_exit=True),
        ])()
      )
  
  def _get_body(self, *args, **kwargs):
    """Build question body and collect answers."""
    answers = list(self.answers.values())

    body = ContentAST.Section()

    body.add_element(
      ContentAST.OnlyHtml([
        ContentAST.Paragraph([
          "Given the following grammar, which of the below strings are part of the language?"
        ])
      ])
    )
    body.add_element(
      ContentAST.OnlyLatex([
        ContentAST.Paragraph([
          "Given the following grammar "
          "please circle any provided strings that are part of the language (or indicate clearly if there are none), "
          "and on each blank line provide generate a new, unique string that is part of the language."
        ])
      ])
    )

    body.add_element(
      ContentAST.Code(self.grammar_good.get_grammar_string())
    )

    # Add in some answers as latex-only options to be circled
    latex_list = ContentAST.OnlyLatex([])
    for answer in self.featured_answers:
      latex_list.add_element(ContentAST.Paragraph([f"- `{str(answer)}`"]))
    body.add_element(latex_list)

    # For Latex-only, ask students to generate some more.
    body.add_element(
      ContentAST.OnlyLatex([
        ContentAST.AnswerBlock([AnswerTypes.String("", label="") for i in range(self.num_answer_blanks)])
      ])
    )

    return body, answers

  def get_body(self, *args, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(*args, **kwargs)
    return body
  
  def _get_explanation(self, *args, **kwargs):
    """Build question explanation."""
    explanation = ContentAST.Section()
    explanation.add_element(
      ContentAST.Paragraph([
        "Remember, for a string to be part of our language, we need to be able to derive it from our grammar.",
        "Unfortunately, there isn't space here to demonstrate the derivation so please work through them on your own!"
      ])
    )
    return explanation, []

  def get_explanation(self, *args, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(*args, **kwargs)
    return explanation

  def get_answers(self, *args, **kwargs) -> Tuple[ContentAST.Answer.CanvasAnswerKind, List[Dict[str,Any]]]:
    
    return ContentAST.Answer.CanvasAnswerKind.MULTIPLE_ANSWER, list(itertools.chain(*[a.get_for_canvas() for a in self.answers.values()]))




import logging.config
import yaml
import os
import re

def setup_logging() -> None:
  config_path = os.path.join(os.path.dirname(__file__), 'logging.yaml')
  if os.path.exists(config_path):
    with open(config_path, 'r') as f:
      config_text = f.read()
    
    # Process environment variables in the format ${VAR:-default}
    def replace_env_vars(match) -> str:
      var_name = match.group(1)
      default_value = match.group(2)
      return os.environ.get(var_name, default_value)
    
    config_text = re.sub(r'\$\{([^}:]+):-([^}]+)\}', replace_env_vars, config_text)
    config = yaml.safe_load(config_text)
    logging.config.dictConfig(config)
  else:
    # Fallback to basic configuration if logging.yaml is not found
    logging.basicConfig(level=logging.INFO)

# Call this once when your application starts
setup_logging()


#!env python
import abc

import numpy as np

from QuizGenerator.question import Question


class MatrixQuestion(Question, abc.ABC):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.default_digits_to_round = kwargs.get("digits_to_round", 2)
  
  def refresh(self, rng_seed=None, *args, **kwargs):
    super().refresh(rng_seed=rng_seed, *args, **kwargs)
    self.rng = np.random.RandomState(rng_seed)
  
  def get_rounded_matrix(self, shape, low=0, high=1, digits_to_round=None):
    if digits_to_round is None:
      digits_to_round = self.default_digits_to_round
    return np.round(
      (high - low) * self.rng.rand(*shape) + low,
      digits_to_round
    )



#!env python
import argparse
from datetime import datetime
import os
import random
import shutil
import subprocess
import tempfile
import traceback
import re
from pathlib import Path
from dotenv import load_dotenv
from QuizGenerator.canvas.canvas_interface import CanvasInterface

from QuizGenerator.quiz import Quiz
from QuizGenerator.question import QuestionRegistry

import logging
log = logging.getLogger(__name__)

from QuizGenerator.performance import PerformanceTracker


def parse_args():
  parser = argparse.ArgumentParser()

  parser.add_argument(
    "--env",
    default=os.path.join(Path.home(), '.env'),
    help="Path to .env file specifying canvas details"
  )
  
  parser.add_argument("--debug", action="store_true", help="Set logging level to debug")

  parser.add_argument("--quiz_yaml", default=os.path.join(os.path.dirname(os.path.abspath(__file__)), "example_files/exam_generation.yaml"))
  parser.add_argument("--seed", type=int, default=None,
                     help="Random seed for quiz generation (default: None for random)")

  # Canvas flags
  parser.add_argument("--num_canvas", default=0, type=int, help="How many variations of each question to try to upload to canvas.")
  parser.add_argument("--prod", action="store_true")
  parser.add_argument("--course_id", type=int)
  parser.add_argument("--delete-assignment-group", action="store_true",
                     help="Delete existing assignment group before uploading new quizzes")
  
  # PDF Flags
  parser.add_argument("--num_pdfs", default=0, type=int, help="How many PDF quizzes to create")
  parser.add_argument("--latex", action="store_false", dest="typst", help="Use Typst instead of LaTeX for PDF generation")

  # Testing flags
  parser.add_argument("--test_all", type=int, default=0, metavar="N",
                     help="Generate N variations of ALL registered questions to test they work correctly")
  parser.add_argument("--strict", action="store_true",
                     help="With --test_all, skip PDF/Canvas generation if any questions fail")

  subparsers = parser.add_subparsers(dest='command')
  test_parser = subparsers.add_parser("TEST")


  args = parser.parse_args()

  if args.num_canvas > 0 and args.course_id is None:
    log.error("Must provide course_id when pushing to canvas")
    exit(8)

  return args


def test():
  log.info("Running test...")

  print("\n" + "="*60)
  print("TEST COMPLETE")
  print("="*60)


def test_all_questions(
    num_variations: int,
    generate_pdf: bool = False,
    use_typst: bool = True,
    canvas_course=None,
    strict: bool = False
):
  """
  Test all registered questions by generating N variations of each.

  This helps verify that all question types work correctly and can generate
  valid output without errors.

  Args:
    num_variations: Number of variations to generate for each question type
    generate_pdf: If True, generate a PDF with all successful questions
    use_typst: If True, use Typst for PDF generation; otherwise use LaTeX
    canvas_course: If provided, push a test quiz to this Canvas course
    strict: If True, skip PDF/Canvas generation if any questions fail
  """
  # Ensure all premade questions are loaded
  QuestionRegistry.load_premade_questions()

  registered_questions = QuestionRegistry._registry
  total_questions = len(registered_questions)

  # Test defaults for questions that require external input
  # These are "template" questions that can't work without content
  TEST_DEFAULTS = {
    'fromtext': {'text': 'Test question placeholder text.'},
    'fromgenerator': {'generator': 'return "Generated test content"'},
  }

  print(f"\nTesting {total_questions} registered question types with {num_variations} variations each...")
  print("=" * 70)

  failed_questions = []
  successful_questions = []
  # Collect question instances for PDF/Canvas generation
  test_question_instances = []

  for i, (question_name, question_class) in enumerate(sorted(registered_questions.items()), 1):
    print(f"\n[{i}/{total_questions}] Testing: {question_name}")
    print(f"  Class: {question_class.__name__}")

    question_failures = []

    for variation in range(num_variations):
      seed = variation * 1000  # Use different seeds for each variation
      try:
        # Get any test defaults for this question type
        extra_kwargs = TEST_DEFAULTS.get(question_name, {})

        # Create question instance with minimal required params
        question = question_class(
          name=f"{question_name} (v{variation+1})",
          points_value=1.0,
          **extra_kwargs
        )

        # Generate the question (this calls refresh and builds the AST)
        question_ast = question.get_question(rng_seed=seed)

        # Try rendering to both formats to catch format-specific issues
        try:
          question_ast.render("html")
        except Exception as e:
          tb = traceback.format_exc()
          question_failures.append(f"  Variation {variation+1}: HTML render failed - {e}\n{tb}")
          continue

        try:
          question_ast.render("typst")
        except Exception as e:
          tb = traceback.format_exc()
          question_failures.append(f"  Variation {variation+1}: Typst render failed - {e}\n{tb}")
          continue

        # If we got here, the question works - save the instance
        test_question_instances.append(question)

      except Exception as e:
        tb = traceback.format_exc()
        question_failures.append(f"  Variation {variation+1}: Generation failed - {e}\n{tb}")

    if question_failures:
      print(f"  FAILED ({len(question_failures)}/{num_variations} variations)")
      for failure in question_failures:
        print(failure)
      failed_questions.append((question_name, question_failures))
    else:
      print(f"  OK ({num_variations}/{num_variations} variations)")
      successful_questions.append(question_name)

  # Summary
  print("\n" + "=" * 70)
  print("TEST SUMMARY")
  print("=" * 70)
  print(f"Total question types: {total_questions}")
  print(f"Successful: {len(successful_questions)}")
  print(f"Failed: {len(failed_questions)}")

  if failed_questions:
    print("\nFailed questions:")
    for name, failures in failed_questions:
      print(f"  - {name}: {len(failures)} failures")

  print("=" * 70)

  # Generate PDF and/or push to Canvas if requested
  if strict and failed_questions:
    print("\n[STRICT MODE] Skipping PDF/Canvas generation due to failures")
  elif (generate_pdf or canvas_course) and test_question_instances:
    print(f"\nCreating test quiz with {len(test_question_instances)} questions...")

    # Create a Quiz object with all successful questions
    test_quiz = Quiz(
      name="Test All Questions",
      questions=test_question_instances,
      practice=True
    )

    if generate_pdf:
      print("Generating PDF...")
      pdf_seed = 12345  # Fixed seed for reproducibility
      if use_typst:
        typst_text = test_quiz.get_quiz(rng_seed=pdf_seed).render("typst")
        generate_typst(typst_text, remove_previous=True, name_prefix="test_all_questions")
      else:
        latex_text = test_quiz.get_quiz(rng_seed=pdf_seed).render_latex()
        generate_latex(latex_text, remove_previous=True, name_prefix="test_all_questions")
      print("PDF generated in out/ directory")

    if canvas_course:
      print("Pushing to Canvas...")
      quiz_title = f"Test All Questions ({int(datetime.now().timestamp())} : {datetime.now().strftime('%b %d %I:%M%p')})"
      canvas_course.push_quiz_to_canvas(
        test_quiz,
        num_variations=1,
        title=quiz_title,
        is_practice=True
      )
      print(f"Quiz '{quiz_title}' pushed to Canvas")

  return len(failed_questions) == 0


def generate_latex(latex_text, remove_previous=False, name_prefix=None):
  """
  Generate PDF from LaTeX source code.

  Args:
    latex_text: The LaTeX source code to compile
    remove_previous: Whether to remove the 'out' directory before generating
    name_prefix: Optional prefix for the temporary filename (e.g., quiz name)
  """
  if remove_previous:
    if os.path.exists('out'): shutil.rmtree('out')

  prefix = f"{sanitize_filename(name_prefix)}-" if name_prefix else "tmp"
  tmp_tex = tempfile.NamedTemporaryFile('w', prefix=prefix)

  tmp_tex.write(latex_text)

  tmp_tex.flush()
  shutil.copy(f"{tmp_tex.name}", "debug.tex")
  p = subprocess.Popen(
    f"latexmk -pdf -output-directory={os.path.join(os.getcwd(), 'out')} {tmp_tex.name}",
    shell=True,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE)
  try:
    p.wait(30)
  except subprocess.TimeoutExpired:
    logging.error("Latex Compile timed out")
    p.kill()
    tmp_tex.close()
    return
  proc = subprocess.Popen(
    f"latexmk -c {tmp_tex.name} -output-directory={os.path.join(os.getcwd(), 'out')}",
    shell=True,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE
  )
  proc.wait(timeout=30)
  tmp_tex.close()


def sanitize_filename(name):
  """
  Sanitize a quiz name for use as a filename prefix.

  Converts spaces to underscores, removes special characters,
  and limits length to avoid overly long filenames.

  Example: "CST 334 Exam 4 (Fall 25)" -> "CST_334_Exam_4_Fall_25"
  """
  # Replace spaces with underscores
  sanitized = name.replace(' ', '_')

  # Remove characters that aren't alphanumeric, underscore, or hyphen
  sanitized = re.sub(r'[^\w\-]', '', sanitized)

  # Limit length to avoid overly long filenames (keep first 50 chars)
  if len(sanitized) > 50:
    sanitized = sanitized[:50]

  return sanitized


def generate_typst(typst_text, remove_previous=False, name_prefix=None):
  """
  Generate PDF from Typst source code.

  Similar to generate_latex, but uses typst compiler instead of latexmk.

  Args:
    typst_text: The Typst source code to compile
    remove_previous: Whether to remove the 'out' directory before generating
    name_prefix: Optional prefix for the temporary filename (e.g., quiz name)
  """
  if remove_previous:
    if os.path.exists('out'):
      shutil.rmtree('out')

  # Ensure output directory exists
  os.makedirs('out', exist_ok=True)

  # Create temporary Typst file with optional name prefix
  prefix = f"{sanitize_filename(name_prefix)}-" if name_prefix else "tmp"
  tmp_typ = tempfile.NamedTemporaryFile('w', suffix='.typ', delete=False, prefix=prefix)

  try:
    tmp_typ.write(typst_text)
    tmp_typ.flush()
    tmp_typ.close()

    # Save debug copy
    shutil.copy(tmp_typ.name, "debug.typ")

    # Compile with typst
    output_pdf = os.path.join(os.getcwd(), 'out', os.path.basename(tmp_typ.name).replace('.typ', '.pdf'))
    
    # Use --root to set the filesystem root so absolute paths work correctly
    p = subprocess.Popen(
      ['typst', 'compile', '--root', '/', tmp_typ.name, output_pdf],
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE
    )

    try:
      p.wait(30)
      if p.returncode != 0:
        stderr = p.stderr.read().decode('utf-8')
        log.error(f"Typst compilation failed: {stderr}")
    except subprocess.TimeoutExpired:
      log.error("Typst compile timed out")
      p.kill()

  finally:
    # Clean up temp file
    if os.path.exists(tmp_typ.name):
      os.unlink(tmp_typ.name)


def generate_quiz(
    path_to_quiz_yaml,
    num_pdfs=0,
    num_canvas=0,
    use_prod=False,
    course_id=None,
    delete_assignment_group=False,
    use_typst=False,
    use_typst_measurement=False,
    base_seed=None
):

  quizzes = Quiz.from_yaml(path_to_quiz_yaml)

  # Handle Canvas uploads with shared assignment group
  if num_canvas > 0:
    canvas_interface = CanvasInterface(prod=use_prod)
    canvas_course = canvas_interface.get_course(course_id=course_id)

    # Create assignment group once, with delete flag if specified
    assignment_group = canvas_course.create_assignment_group(
      name="dev",
      delete_existing=delete_assignment_group
    )

    log.info(f"Using assignment group '{assignment_group.name}' for all quizzes")

  for quiz in quizzes:

    for i in range(num_pdfs):
      log.debug(f"Generating PDF {i+1}/{num_pdfs}")
      # If base_seed is provided, use it with an offset for each PDF
      # Otherwise generate a random seed for this PDF
      if base_seed is not None:
        pdf_seed = base_seed + (i * 1000)  # Large gap to avoid overlap with rng_seed_offset
      else:
        pdf_seed = random.randint(0, 1_000_000)

      log.info(f"Generating PDF {i+1} with seed: {pdf_seed}")

      if use_typst:
        # Generate using Typst
        typst_text = quiz.get_quiz(rng_seed=pdf_seed, use_typst_measurement=use_typst_measurement).render("typst")
        generate_typst(typst_text, remove_previous=(i==0), name_prefix=quiz.name)
      else:
        # Generate using LaTeX (default)
        latex_text = quiz.get_quiz(rng_seed=pdf_seed, use_typst_measurement=use_typst_measurement).render_latex()
        generate_latex(latex_text, remove_previous=(i==0), name_prefix=quiz.name)

    if num_canvas > 0:
      canvas_course.push_quiz_to_canvas(
        quiz,
        num_canvas,
        title=quiz.name,
        is_practice=quiz.practice,
        assignment_group=assignment_group
      )
    
    quiz.describe()

def main():

  args = parse_args()
  
  # Load environment variables
  load_dotenv(args.env)
  
  if args.debug:
    # Set root logger to DEBUG
    logging.getLogger().setLevel(logging.DEBUG)

    # Set all handlers to DEBUG level
    for handler in logging.getLogger().handlers:
      handler.setLevel(logging.DEBUG)

    # Set named loggers to DEBUG
    for logger_name in ['QuizGenerator', 'lms_interface', '__main__']:
      logger = logging.getLogger(logger_name)
      logger.setLevel(logging.DEBUG)
      for handler in logger.handlers:
        handler.setLevel(logging.DEBUG)

  if args.command == "TEST":
    test()
    return

  if args.test_all > 0:
    # Set up Canvas course if course_id provided
    canvas_course = None
    if args.course_id:
      canvas_interface = CanvasInterface(prod=args.prod)
      canvas_course = canvas_interface.get_course(course_id=args.course_id)

    success = test_all_questions(
      args.test_all,
      generate_pdf=True,
      use_typst=getattr(args, 'typst', True),
      canvas_course=canvas_course,
      strict=args.strict
    )
    exit(0 if success else 1)

  # Clear any previous metrics
  PerformanceTracker.clear_metrics()

  generate_quiz(
    args.quiz_yaml,
    num_pdfs=args.num_pdfs,
    num_canvas=args.num_canvas,
    use_prod=args.prod,
    course_id=args.course_id,
    delete_assignment_group=getattr(args, 'delete_assignment_group', False),
    use_typst=getattr(args, 'typst', False),
    use_typst_measurement=getattr(args, 'typst_measurement', False),
    base_seed=getattr(args, 'seed', None)
  )


if __name__ == "__main__":
  main()



from __future__ import annotations

import abc
import io
import logging
import re
import numpy as np
import sympy as sp
from typing import List, Tuple, Dict, Any

from QuizGenerator.contentast import ContentAST, AnswerTypes
from QuizGenerator.question import Question, QuestionRegistry
from QuizGenerator.mixins import TableQuestionMixin, BodyTemplatesMixin

# Import gradient descent utilities
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'gradient_descent'))
from misc import generate_function, format_vector

log = logging.getLogger(__name__)


@QuestionRegistry.register()
class ParameterCountingQuestion(Question):
  """
  Question asking students to count parameters in a neural network.

  Given a dense network architecture, students calculate:
  - Total number of weights
  - Total number of biases
  - Total trainable parameters
  """

  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.ML_OPTIMIZATION)
    super().__init__(*args, **kwargs)

    self.num_layers = kwargs.get("num_layers", None)
    self.include_biases = kwargs.get("include_biases", True)

  def refresh(self, rng_seed=None, *args, **kwargs):
    super().refresh(rng_seed=rng_seed, *args, **kwargs)

    # Generate random architecture
    if self.num_layers is None:
      self.num_layers = self.rng.choice([3, 4])

    # Generate layer sizes
    # Input layer: common sizes for typical problems
    input_sizes = [28*28, 32*32, 784, 1024, 64, 128]
    self.layer_sizes = [self.rng.choice(input_sizes)]

    # Hidden layers: reasonable sizes
    for i in range(self.num_layers - 2):
      hidden_size = self.rng.choice([32, 64, 128, 256, 512])
      self.layer_sizes.append(hidden_size)

    # Output layer: typical classification sizes
    output_size = self.rng.choice([2, 10, 100, 1000])
    self.layer_sizes.append(output_size)

    # Calculate correct answers
    self.total_weights = 0
    self.total_biases = 0
    self.weights_per_layer = []
    self.biases_per_layer = []

    for i in range(len(self.layer_sizes) - 1):
      weights = self.layer_sizes[i] * self.layer_sizes[i+1]
      biases = self.layer_sizes[i+1] if self.include_biases else 0

      self.weights_per_layer.append(weights)
      self.biases_per_layer.append(biases)

      self.total_weights += weights
      self.total_biases += biases

    self.total_params = self.total_weights + self.total_biases

    # Create answers
    self._create_answers()

  def _create_answers(self):
    """Create answer fields."""
    self.answers = {}

    self.answers["total_weights"] = AnswerTypes.Int(self.total_weights, label="Total weights")

    if self.include_biases:
      self.answers["total_biases"] = AnswerTypes.Int(self.total_biases, label="Total biases")
      self.answers["total_params"] = AnswerTypes.Int(self.total_params, label="Total trainable parameters")
    else:
      self.answers["total_params"] = AnswerTypes.Int(self.total_params, label="Total trainable parameters")

  def _get_body(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question body and collect answers."""
    body = ContentAST.Section()
    answers = []

    # Question description
    body.add_element(ContentAST.Paragraph([
      "Consider a fully-connected (dense) neural network with the following architecture:"
    ]))

    # Display architecture
    arch_parts = []
    for i, size in enumerate(self.layer_sizes):
      if i > 0:
        arch_parts.append(" → ")
      arch_parts.append(str(size))

    body.add_element(ContentAST.Paragraph([
      "Architecture: " + "".join(arch_parts)
    ]))

    if self.include_biases:
      body.add_element(ContentAST.Paragraph([
        "Each layer includes bias terms."
      ]))

    # Questions
    # Answer table
    table_data = []
    table_data.append(["Parameter Type", "Count"])

    answers.append(self.answers["total_weights"])
    table_data.append([
      "Total weights (connections between layers)",
      self.answers["total_weights"]
    ])

    if self.include_biases:
      answers.append(self.answers["total_biases"])
      table_data.append([
        "Total biases",
        self.answers["total_biases"]
      ])

    answers.append(self.answers["total_params"])
    table_data.append([
      "Total trainable parameters",
      self.answers["total_params"]
    ])

    body.add_element(ContentAST.Table(data=table_data))

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body

  def _get_explanation(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph([
      "To count parameters in a dense neural network, we calculate weights and biases for each layer."
    ]))

    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Weights calculation:", emphasis=True)
    ]))

    for i in range(len(self.layer_sizes) - 1):
      input_size = self.layer_sizes[i]
      output_size = self.layer_sizes[i+1]
      weights = self.weights_per_layer[i]

      explanation.add_element(ContentAST.Paragraph([
        f"Layer {i+1} → {i+2}: ",
        ContentAST.Equation(f"{input_size} \\times {output_size} = {weights:,}", inline=True),
        " weights"
      ]))

    explanation.add_element(ContentAST.Paragraph([
      "Total weights: ",
      ContentAST.Equation(
        f"{' + '.join([f'{w:,}' for w in self.weights_per_layer])} = {self.total_weights:,}",
        inline=True
      )
    ]))

    if self.include_biases:
      explanation.add_element(ContentAST.Paragraph([
        ContentAST.Text("Biases calculation:", emphasis=True)
      ]))

      for i in range(len(self.layer_sizes) - 1):
        output_size = self.layer_sizes[i+1]
        biases = self.biases_per_layer[i]

        explanation.add_element(ContentAST.Paragraph([
          f"Layer {i+2}: {biases:,} biases (one per neuron)"
        ]))

      explanation.add_element(ContentAST.Paragraph([
        "Total biases: ",
        ContentAST.Equation(
          f"{' + '.join([f'{b:,}' for b in self.biases_per_layer])} = {self.total_biases:,}",
          inline=True
        )
      ]))

    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Total trainable parameters:", emphasis=True)
    ]))

    if self.include_biases:
      explanation.add_element(ContentAST.Equation(
        f"\\text{{Total}} = {self.total_weights:,} + {self.total_biases:,} = {self.total_params:,}",
        inline=False
      ))
    else:
      explanation.add_element(ContentAST.Equation(
        f"\\text{{Total}} = {self.total_weights:,}",
        inline=False
      ))

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation


@QuestionRegistry.register()
class ActivationFunctionComputationQuestion(Question):
  """
  Question asking students to compute activation function outputs.

  Given a vector of inputs and an activation function, students calculate
  the output for each element (or entire vector for softmax).
  """

  ACTIVATION_RELU = "relu"
  ACTIVATION_SIGMOID = "sigmoid"
  ACTIVATION_TANH = "tanh"
  ACTIVATION_SOFTMAX = "softmax"

  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.ML_OPTIMIZATION)
    super().__init__(*args, **kwargs)

    self.vector_size = kwargs.get("vector_size", None)
    self.activation = kwargs.get("activation", None)

  def refresh(self, rng_seed=None, *args, **kwargs):
    super().refresh(rng_seed=rng_seed, *args, **kwargs)

    # Generate random input vector
    if self.vector_size is None:
      self.vector_size = self.rng.choice([3, 4, 5])

    self.input_vector = [
      round(self.rng.uniform(-3, 3), 1)
      for _ in range(self.vector_size)
    ]

    # Select activation function
    if self.activation is None:
      activations = [
        self.ACTIVATION_RELU,
        self.ACTIVATION_SIGMOID,
        self.ACTIVATION_TANH,
        self.ACTIVATION_SOFTMAX,
      ]
      self.activation = self.rng.choice(activations)

    # For leaky ReLU, set alpha
    self.leaky_alpha = 0.01

    # Compute outputs
    self.output_vector = self._compute_activation(self.input_vector)

    # Create answers
    self._create_answers()

  def _compute_activation(self, inputs):
    """Compute activation function output."""
    if self.activation == self.ACTIVATION_RELU:
      return [max(0, x) for x in inputs]

    elif self.activation == self.ACTIVATION_SIGMOID:
      return [1 / (1 + np.exp(-x)) for x in inputs]

    elif self.activation == self.ACTIVATION_TANH:
      return [np.tanh(x) for x in inputs]

    elif self.activation == self.ACTIVATION_SOFTMAX:
      # Subtract max for numerical stability
      exp_vals = [np.exp(x - max(inputs)) for x in inputs]
      sum_exp = sum(exp_vals)
      return [e / sum_exp for e in exp_vals]

    else:
      raise ValueError(f"Unknown activation: {self.activation}")

  def _get_activation_name(self):
    """Get human-readable activation name."""
    names = {
      self.ACTIVATION_RELU: "ReLU",
      self.ACTIVATION_SIGMOID: "Sigmoid",
      self.ACTIVATION_TANH: "Tanh",
      self.ACTIVATION_SOFTMAX: "Softmax",
    }
    return names.get(self.activation, "Unknown")

  def _get_activation_formula(self):
    """Get LaTeX formula for activation function."""
    if self.activation == self.ACTIVATION_RELU:
      return r"\text{ReLU}(x) = \max(0, x)"

    elif self.activation == self.ACTIVATION_SIGMOID:
      return r"\sigma(x) = \frac{1}{1 + e^{-x}}"

    elif self.activation == self.ACTIVATION_TANH:
      return r"\tanh(x) = \frac{e^x - e^{-x}}{e^x + e^{-x}}"

    elif self.activation == self.ACTIVATION_SOFTMAX:
      return r"\text{softmax}(x_i) = \frac{e^{x_i}}{\sum_j e^{x_j}}"

    return ""

  def _create_answers(self):
    """Create answer fields."""
    self.answers = {}

    if self.activation == self.ACTIVATION_SOFTMAX:
      # Softmax: single vector answer
      self.answers["output"] = AnswerTypes.Vector(self.output_vector, label="Output vector")
    else:
      # Element-wise: individual answers
      for i, output in enumerate(self.output_vector):
        key = f"output_{i}"
        self.answers[key] = AnswerTypes.Float(float(output), label=f"Output for input {self.input_vector[i]:.1f}")

  def _get_body(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question body and collect answers."""
    body = ContentAST.Section()
    answers = []

    # Question description
    body.add_element(ContentAST.Paragraph([
      f"Given the input vector below, compute the output after applying the {self._get_activation_name()} activation function."
    ]))

    # Display formula
    body.add_element(ContentAST.Paragraph([
      "Activation function: ",
      ContentAST.Equation(self._get_activation_formula(), inline=True)
    ]))

    # Input vector
    input_str = ", ".join([f"{x:.1f}" for x in self.input_vector])
    body.add_element(ContentAST.Paragraph([
      "Input: ",
      ContentAST.Equation(f"[{input_str}]", inline=True)
    ]))

    # Answer table
    if self.activation == self.ACTIVATION_SOFTMAX:
      body.add_element(ContentAST.Paragraph([
        "Compute the output vector:"
      ]))

      answers.append(self.answers["output"])
      table_data = []
      table_data.append(["Output Vector"])
      table_data.append([self.answers["output"]])

      body.add_element(ContentAST.Table(data=table_data))

    else:
      body.add_element(ContentAST.Paragraph([
        "Compute the output for each element:"
      ]))

      table_data = []
      table_data.append(["Input", "Output"])

      for i, x in enumerate(self.input_vector):
        answer = self.answers[f"output_{i}"]
        answers.append(answer)
        table_data.append([
          ContentAST.Equation(f"{x:.1f}", inline=True),
          answer
        ])

      body.add_element(ContentAST.Table(data=table_data))

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body

  def _get_explanation(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph([
      f"To compute the {self._get_activation_name()} activation, we apply the formula to each input."
    ]))

    if self.activation == self.ACTIVATION_SOFTMAX:
      explanation.add_element(ContentAST.Paragraph([
        ContentAST.Text("Softmax computation:", emphasis=True)
      ]))

      # Show exponentials
      exp_strs = [f"e^{{{x:.1f}}}" for x in self.input_vector]
      explanation.add_element(ContentAST.Paragraph([
        "First, compute exponentials: ",
        ContentAST.Equation(", ".join(exp_strs), inline=True)
      ]))

      # Numerical values
      exp_vals = [np.exp(x) for x in self.input_vector]
      exp_vals_str = ", ".join([f"{e:.4f}" for e in exp_vals])
      explanation.add_element(ContentAST.Paragraph([
        ContentAST.Equation(f"\\approx [{exp_vals_str}]", inline=True)
      ]))

      # Sum
      sum_exp = sum(exp_vals)
      explanation.add_element(ContentAST.Paragraph([
        "Sum: ",
        ContentAST.Equation(f"{sum_exp:.4f}", inline=True)
      ]))

      # Final outputs
      explanation.add_element(ContentAST.Paragraph([
        "Divide each by the sum:"
      ]))

      for i, (exp_val, output) in enumerate(zip(exp_vals, self.output_vector)):
        explanation.add_element(ContentAST.Equation(
          f"\\text{{softmax}}({self.input_vector[i]:.1f}) = \\frac{{{exp_val:.4f}}}{{{sum_exp:.4f}}} = {output:.4f}",
          inline=False
        ))

    else:
      explanation.add_element(ContentAST.Paragraph([
        ContentAST.Text("Element-wise computation:", emphasis=True)
      ]))

      for i, (x, y) in enumerate(zip(self.input_vector, self.output_vector)):
        if self.activation == self.ACTIVATION_RELU:
          explanation.add_element(ContentAST.Equation(
            f"\\text{{ReLU}}({x:.1f}) = \\max(0, {x:.1f}) = {y:.4f}",
            inline=False
          ))

        elif self.activation == self.ACTIVATION_SIGMOID:
          explanation.add_element(ContentAST.Equation(
            f"\\sigma({x:.1f}) = \\frac{{1}}{{1 + e^{{-{x:.1f}}}}} = {y:.4f}",
            inline=False
          ))

        elif self.activation == self.ACTIVATION_TANH:
          explanation.add_element(ContentAST.Equation(
            f"\\tanh({x:.1f}) = {y:.4f}",
            inline=False
          ))

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation


@QuestionRegistry.register()
class RegularizationCalculationQuestion(Question):
  """
  Question asking students to calculate loss with L2 regularization.

  Given a small network (2-4 weights), students calculate:
  - Forward pass
  - Base MSE loss
  - L2 regularization penalty
  - Total loss
  - Gradient with regularization for one weight
  """

  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.ML_OPTIMIZATION)
    super().__init__(*args, **kwargs)

    self.num_weights = kwargs.get("num_weights", None)

  def refresh(self, rng_seed=None, *args, **kwargs):
    super().refresh(rng_seed=rng_seed, *args, **kwargs)

    # Generate small network (2-4 weights for simplicity)
    if self.num_weights is None:
      self.num_weights = self.rng.choice([2, 3, 4])

    # Generate weights (small values)
    self.weights = [
      round(self.rng.uniform(-2, 2), 1)
      for _ in range(self.num_weights)
    ]

    # Generate input and target
    self.input_val = round(self.rng.uniform(-3, 3), 1)
    self.target = round(self.rng.uniform(-5, 5), 1)

    # Regularization coefficient
    self.lambda_reg = self.rng.choice([0.01, 0.05, 0.1, 0.5])

    # Forward pass (simple linear combination for simplicity)
    # prediction = sum(w_i * input^i) for i in 0..n
    # This gives us a polynomial: w0 + w1*x + w2*x^2 + ...
    self.prediction = sum(
      w * (self.input_val ** i)
      for i, w in enumerate(self.weights)
    )

    # Calculate losses
    self.base_loss = 0.5 * (self.target - self.prediction) ** 2
    self.l2_penalty = (self.lambda_reg / 2) * sum(w**2 for w in self.weights)
    self.total_loss = self.base_loss + self.l2_penalty

    # Calculate gradient for first weight (w0, the bias term)
    # dL_base/dw0 = -(target - prediction) * dPrediction/dw0
    # dPrediction/dw0 = input^0 = 1
    # dL_reg/dw0 = lambda * w0
    # dL_total/dw0 = dL_base/dw0 + dL_reg/dw0

    self.grad_base_w0 = -(self.target - self.prediction) * 1  # derivative of w0*x^0
    self.grad_reg_w0 = self.lambda_reg * self.weights[0]
    self.grad_total_w0 = self.grad_base_w0 + self.grad_reg_w0

    # Create answers
    self._create_answers()

  def _create_answers(self):
    """Create answer fields."""
    self.answers = {}

    self.answers["prediction"] = AnswerTypes.Float(float(self.prediction), label="Prediction ŷ")
    self.answers["base_loss"] = AnswerTypes.Float(float(self.base_loss), label="Base MSE loss")
    self.answers["l2_penalty"] = AnswerTypes.Float(float(self.l2_penalty), label="L2 penalty")
    self.answers["total_loss"] = AnswerTypes.Float(float(self.total_loss), label="Total loss")
    self.answers["grad_total_w0"] = AnswerTypes.Float(float(self.grad_total_w0), label="Gradient ∂L/∂w₀")

  def _get_body(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question body and collect answers."""
    body = ContentAST.Section()
    answers = []

    # Question description
    body.add_element(ContentAST.Paragraph([
      "Consider a simple model with the following parameters:"
    ]))

    # Display weights
    weight_strs = [f"w_{i} = {w:.1f}" for i, w in enumerate(self.weights)]
    body.add_element(ContentAST.Paragraph([
      "Weights: ",
      ContentAST.Equation(", ".join(weight_strs), inline=True)
    ]))

    # Model equation
    terms = []
    for i, w in enumerate(self.weights):
      if i == 0:
        terms.append(f"w_0")
      elif i == 1:
        terms.append(f"w_1 x")
      else:
        terms.append(f"w_{i} x^{i}")

    model_eq = " + ".join(terms)
    body.add_element(ContentAST.Paragraph([
      "Model: ",
      ContentAST.Equation(f"\\hat{{y}} = {model_eq}", inline=True)
    ]))

    # Data point
    body.add_element(ContentAST.Paragraph([
      "Data point: ",
      ContentAST.Equation(f"x = {self.input_val:.1f}, y = {self.target:.1f}", inline=True)
    ]))

    # Regularization
    body.add_element(ContentAST.Paragraph([
      "L2 regularization coefficient: ",
      ContentAST.Equation(f"\\lambda = {self.lambda_reg}", inline=True)
    ]))

    body.add_element(ContentAST.Paragraph([
      "Calculate the following:"
    ]))

    # Answer table
    table_data = []
    table_data.append(["Calculation", "Value"])

    answers.append(self.answers["prediction"])
    table_data.append([
      ContentAST.Paragraph(["Prediction ", ContentAST.Equation(r"\hat{y}", inline=True)]),
      self.answers["prediction"]
    ])

    answers.append(self.answers["base_loss"])
    table_data.append([
      ContentAST.Paragraph(["Base MSE loss: ", ContentAST.Equation(r"L_{base} = (1/2)(y - \hat{y})^2", inline=True)]),
      self.answers["base_loss"]
    ])

    answers.append(self.answers["l2_penalty"])
    table_data.append([
      ContentAST.Paragraph(["L2 penalty: ", ContentAST.Equation(r"L_{reg} = (\lambda/2)\sum w_i^2", inline=True)]),
      self.answers["l2_penalty"]
    ])

    answers.append(self.answers["total_loss"])
    table_data.append([
      ContentAST.Paragraph(["Total loss: ", ContentAST.Equation(r"L_{total} = L_{base} + L_{reg}", inline=True)]),
      self.answers["total_loss"]
    ])

    answers.append(self.answers["grad_total_w0"])
    table_data.append([
      ContentAST.Paragraph(["Gradient: ", ContentAST.Equation(r"\frac{\partial L_{total}}{\partial w_0}", inline=True)]),
      self.answers["grad_total_w0"]
    ])

    body.add_element(ContentAST.Table(data=table_data))

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body

  def _get_explanation(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph([
      "L2 regularization adds a penalty term to the loss function to prevent overfitting by keeping weights small."
    ]))

    # Step 1: Forward pass
    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Step 1: Compute prediction", emphasis=True)
    ]))

    terms = []
    for i, w in enumerate(self.weights):
      if i == 0:
        terms.append(f"{w:.1f}")
      else:
        x_term = f"{self.input_val:.1f}^{i}" if i > 1 else f"{self.input_val:.1f}"
        terms.append(f"{w:.1f} \\times {x_term}")

    explanation.add_element(ContentAST.Equation(
      f"\\hat{{y}} = {' + '.join(terms)} = {self.prediction:.4f}",
      inline=False
    ))

    # Step 2: Base loss
    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Step 2: Compute base MSE loss", emphasis=True)
    ]))

    explanation.add_element(ContentAST.Equation(
      f"L_{{base}} = \\frac{{1}}{{2}}(y - \\hat{{y}})^2 = \\frac{{1}}{{2}}({self.target:.1f} - {self.prediction:.4f})^2 = {self.base_loss:.4f}",
      inline=False
    ))

    # Step 3: L2 penalty
    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Step 3: Compute L2 penalty", emphasis=True)
    ]))

    weight_squares = [f"{w:.1f}^2" for w in self.weights]
    sum_squares = sum(w**2 for w in self.weights)

    explanation.add_element(ContentAST.Equation(
      f"L_{{reg}} = \\frac{{\\lambda}}{{2}} \\sum w_i^2 = \\frac{{{self.lambda_reg}}}{{2}}({' + '.join(weight_squares)}) = \\frac{{{self.lambda_reg}}}{{2}} \\times {sum_squares:.4f} = {self.l2_penalty:.4f}",
      inline=False
    ))

    # Step 4: Total loss
    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Step 4: Compute total loss", emphasis=True)
    ]))

    explanation.add_element(ContentAST.Equation(
      f"L_{{total}} = L_{{base}} + L_{{reg}} = {self.base_loss:.4f} + {self.l2_penalty:.4f} = {self.total_loss:.4f}",
      inline=False
    ))

    # Step 5: Gradient with regularization
    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Step 5: Compute gradient with regularization", emphasis=True)
    ]))

    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Equation(r"w_0", inline=True),
      " (the bias term):"
    ]))

    explanation.add_element(ContentAST.Equation(
      f"\\frac{{\\partial L_{{base}}}}{{\\partial w_0}} = -(y - \\hat{{y}}) \\times 1 = -({self.target:.1f} - {self.prediction:.4f}) = {self.grad_base_w0:.4f}",
      inline=False
    ))

    explanation.add_element(ContentAST.Equation(
      f"\\frac{{\\partial L_{{reg}}}}{{\\partial w_0}} = \\lambda w_0 = {self.lambda_reg} \\times {self.weights[0]:.1f} = {self.grad_reg_w0:.4f}",
      inline=False
    ))

    explanation.add_element(ContentAST.Equation(
      f"\\frac{{\\partial L_{{total}}}}{{\\partial w_0}} = {self.grad_base_w0:.4f} + {self.grad_reg_w0:.4f} = {self.grad_total_w0:.4f}",
      inline=False
    ))

    explanation.add_element(ContentAST.Paragraph([
      "The regularization term adds ",
      ContentAST.Equation(f"\\lambda w_0 = {self.grad_reg_w0:.4f}", inline=True),
      " to the gradient, pushing the weight toward zero."
    ]))

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation


@QuestionRegistry.register()
class MomentumOptimizerQuestion(Question, TableQuestionMixin, BodyTemplatesMixin):
  """
  Question asking students to perform gradient descent with momentum.

  Given a function, current weights, gradients, learning rate, and momentum coefficient,
  students calculate:
  - Velocity update using momentum
  - Weight update using the new velocity
  - Comparison to vanilla SGD (optional)
  """

  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.ML_OPTIMIZATION)
    super().__init__(*args, **kwargs)

    self.num_variables = kwargs.get("num_variables", 2)
    self.show_vanilla_sgd = kwargs.get("show_vanilla_sgd", True)

  def refresh(self, rng_seed=None, *args, **kwargs):
    super().refresh(rng_seed=rng_seed, *args, **kwargs)

    # Generate well-conditioned quadratic function
    self.variables, self.function, self.gradient_function, self.equation = \
        generate_function(self.rng, self.num_variables, max_degree=2, use_quadratic=True)

    # Generate current weights (small integers)
    self.current_weights = [
      self.rng.choice([-2, -1, 0, 1, 2])
      for _ in range(self.num_variables)
    ]

    # Calculate gradient at current position
    subs_map = dict(zip(self.variables, self.current_weights))
    g_syms = self.gradient_function.subs(subs_map)
    self.gradients = [float(val) for val in g_syms]

    # Generate previous velocity (for momentum)
    # Start with small or zero velocity
    self.prev_velocity = [
      round(self.rng.uniform(-0.5, 0.5), 2)
      for _ in range(self.num_variables)
    ]

    # Hyperparameters
    self.learning_rate = self.rng.choice([0.01, 0.05, 0.1])
    self.momentum_beta = self.rng.choice([0.8, 0.9])

    # Calculate momentum updates
    # v_new = beta * v_old + (1 - beta) * gradient
    self.new_velocity = [
      self.momentum_beta * v_old + (1 - self.momentum_beta) * grad
      for v_old, grad in zip(self.prev_velocity, self.gradients)
    ]

    # w_new = w_old - alpha * v_new
    self.new_weights = [
      w - self.learning_rate * v
      for w, v in zip(self.current_weights, self.new_velocity)
    ]

    # Calculate vanilla SGD for comparison
    if self.show_vanilla_sgd:
      self.sgd_weights = [
        w - self.learning_rate * grad
        for w, grad in zip(self.current_weights, self.gradients)
      ]

    # Create answers
    self._create_answers()

  def _create_answers(self):
    """Create answer fields."""
    self.answers = {}

    # New velocity
    self.answers["velocity"] = AnswerTypes.Vector(self.new_velocity, label="New velocity")

    # New weights with momentum
    self.answers["weights_momentum"] = AnswerTypes.Vector(self.new_weights, label="Weights (momentum)")

    # Vanilla SGD weights for comparison
    if self.show_vanilla_sgd:
      self.answers["weights_sgd"] = AnswerTypes.Vector(self.sgd_weights, label="Weights (vanilla SGD)")

  def _get_body(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question body and collect answers."""
    body = ContentAST.Section()
    answers = []

    # Question description
    body.add_element(ContentAST.Paragraph([
      "Consider the optimization problem of minimizing the function:"
    ]))

    body.add_element(ContentAST.Equation(
      sp.latex(self.function),
      inline=False
    ))

    body.add_element(ContentAST.Paragraph([
      "The gradient is:"
    ]))

    body.add_element(ContentAST.Equation(
      f"\\nabla f = {sp.latex(self.gradient_function)}",
      inline=False
    ))

    # Current state
    body.add_element(ContentAST.Paragraph([
      ContentAST.Text("Current optimization state:", emphasis=True)
    ]))

    body.add_element(ContentAST.Paragraph([
      "Current weights: ",
      ContentAST.Equation(f"{format_vector(self.current_weights)}", inline=True)
    ]))

    body.add_element(ContentAST.Paragraph([
      "Previous velocity: ",
      ContentAST.Equation(f"{format_vector(self.prev_velocity)}", inline=True)
    ]))

    # Hyperparameters
    body.add_element(ContentAST.Paragraph([
      ContentAST.Text("Hyperparameters:", emphasis=True)
    ]))

    body.add_element(ContentAST.Paragraph([
      "Learning rate: ",
      ContentAST.Equation(f"\\alpha = {self.learning_rate}", inline=True)
    ]))

    body.add_element(ContentAST.Paragraph([
      "Momentum coefficient: ",
      ContentAST.Equation(f"\\beta = {self.momentum_beta}", inline=True)
    ]))

    # Questions
    body.add_element(ContentAST.Paragraph([
      "Calculate the following updates:"
    ]))

    # Answer table
    table_data = []
    table_data.append(["Update Type", "Formula", "Result"])

    answers.append(self.answers["velocity"])
    table_data.append([
      "New velocity",
      ContentAST.Equation(r"v' = \beta v + (1-\beta)\nabla f", inline=True),
      self.answers["velocity"]
    ])

    answers.append(self.answers["weights_momentum"])
    table_data.append([
      "Weights (momentum)",
      ContentAST.Equation(r"w' = w - \alpha v'", inline=True),
      self.answers["weights_momentum"]
    ])

    if self.show_vanilla_sgd:
      answers.append(self.answers["weights_sgd"])
      table_data.append([
        "Weights (vanilla SGD)",
        ContentAST.Equation(r"w' = w - \alpha \nabla f", inline=True),
        self.answers["weights_sgd"]
      ])

    body.add_element(ContentAST.Table(data=table_data))

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body

  def _get_explanation(self, **kwargs) -> Tuple[ContentAST.Section, List[ContentAST.Answer]]:
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph([
      "Momentum helps gradient descent by accumulating a velocity vector in directions of "
      "consistent gradient, allowing faster convergence and reduced oscillation."
    ]))

    # Step 1: Calculate new velocity
    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Step 1: Update velocity using momentum", emphasis=True)
    ]))

    explanation.add_element(ContentAST.Paragraph([
      "The momentum update formula is:"
    ]))

    explanation.add_element(ContentAST.Equation(
      f"v' = \\beta v + (1 - \\beta) \\nabla f",
      inline=False
    ))

    # Show calculation for each component
    digits = ContentAST.Answer.DEFAULT_ROUNDING_DIGITS
    for i in range(self.num_variables):
      var_name = f"x_{i}"
      # Round all intermediate values to avoid floating point precision issues
      beta_times_v = round(self.momentum_beta * self.prev_velocity[i], digits)
      one_minus_beta = round(1 - self.momentum_beta, digits)
      one_minus_beta_times_grad = round((1 - self.momentum_beta) * self.gradients[i], digits)

      explanation.add_element(ContentAST.Equation(
        f"v'[{i}] = {self.momentum_beta} \\times {self.prev_velocity[i]:.{digits}f} + "
        f"{one_minus_beta:.{digits}f} \\times {self.gradients[i]:.{digits}f} = "
        f"{beta_times_v:.{digits}f} + {one_minus_beta_times_grad:.{digits}f} = {self.new_velocity[i]:.{digits}f}",
        inline=False
      ))

    # Step 2: Update weights with momentum
    explanation.add_element(ContentAST.Paragraph([
      ContentAST.Text("Step 2: Update weights using new velocity", emphasis=True)
    ]))

    explanation.add_element(ContentAST.Equation(
      f"w' = w - \\alpha v'",
      inline=False
    ))

    for i in range(self.num_variables):
      explanation.add_element(ContentAST.Equation(
        f"w[{i}] = {self.current_weights[i]} - {self.learning_rate} \\times {self.new_velocity[i]:.4f} = {self.new_weights[i]:.4f}",
        inline=False
      ))

    # Comparison with vanilla SGD
    if self.show_vanilla_sgd:
      explanation.add_element(ContentAST.Paragraph([
        ContentAST.Text("Comparison with vanilla SGD:", emphasis=True)
      ]))

      explanation.add_element(ContentAST.Paragraph([
        "Vanilla SGD (no momentum) would update directly using the gradient:"
      ]))

      explanation.add_element(ContentAST.Equation(
        f"w' = w - \\alpha \\nabla f",
        inline=False
      ))

      for i in range(self.num_variables):
        explanation.add_element(ContentAST.Equation(
          f"w[{i}] = {self.current_weights[i]} - {self.learning_rate} \\times {self.gradients[i]:.4f} = {self.sgd_weights[i]:.4f}",
          inline=False
        ))

      explanation.add_element(ContentAST.Paragraph([
        "The momentum update differs because it incorporates the previous velocity, "
        "which can help accelerate learning and smooth out noisy gradients."
      ]))

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation



#!env python
import abc
import logging
import math

from QuizGenerator.question import Question, QuestionRegistry
from QuizGenerator.contentast import ContentAST, AnswerTypes
from QuizGenerator.constants import MathRanges

log = logging.getLogger(__name__)


class MathQuestion(Question, abc.ABC):
  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.MATH)
    super().__init__(*args, **kwargs)


@QuestionRegistry.register()
class BitsAndBytes(MathQuestion):
  
  MIN_BITS = MathRanges.DEFAULT_MIN_MATH_BITS
  MAX_BITS = MathRanges.DEFAULT_MAX_MATH_BITS
  
  def refresh(self, *args, **kwargs):
    super().refresh(*args, **kwargs)
    
    # Generate the important parts of the problem
    self.from_binary = (0 == self.rng.randint(0,1))
    self.num_bits = self.rng.randint(self.MIN_BITS, self.MAX_BITS)
    self.num_bytes = int(math.pow(2, self.num_bits))
    
    if self.from_binary:
      self.answers = {"answer": AnswerTypes.Int(self.num_bytes,
                                               label="Address space size", unit="Bytes")}
    else:
      self.answers = {"answer": AnswerTypes.Int(self.num_bits,
                                               label="Number of bits in address", unit="bits")}
  
  def _get_body(self, **kwargs):
    """Build question body and collect answers."""
    answers = [self.answers['answer']]

    body = ContentAST.Section()
    body.add_element(
      ContentAST.Paragraph([
        f"Given that we have "
        f"{self.num_bits if self.from_binary else self.num_bytes} {'bits' if self.from_binary else 'bytes'}, "
        f"how many {'bits' if not self.from_binary else 'bytes'} "
        f"{'do we need to address our memory' if not self.from_binary else 'of memory can be addressed'}?"
      ])
    )

    body.add_element(ContentAST.AnswerBlock(self.answers['answer']))

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body

  def _get_explanation(self, **kwargs):
    explanation = ContentAST.Section()
    
    explanation.add_element(
      ContentAST.Paragraph([
        "Remember that for these problems we use one of these two equations (which are equivalent)"
      ])
    )
    explanation.add_elements([
      ContentAST.Equation(r"log_{2}(\text{#bytes}) = \text{#bits}"),
      ContentAST.Equation(r"2^{(\text{#bits})} = \text{#bytes}")
    ])
    
    explanation.add_element(
      ContentAST.Paragraph(["Therefore, we calculate:"])
    )
    
    if self.from_binary:
      explanation.add_element(
        ContentAST.Equation(f"2 ^ {{{self.num_bits}bits}} = \\textbf{{{self.num_bytes}}}\\text{{bytes}}")
      )
    else:
      explanation.add_element(
        ContentAST.Equation(f"log_{{2}}({self.num_bytes} \\text{{bytes}}) = \\textbf{{{self.num_bits}}}\\text{{bits}}")
      )

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation


@QuestionRegistry.register()
class HexAndBinary(MathQuestion):
  
  MIN_HEXITS = 1
  MAX_HEXITS = 8
  
  def refresh(self, **kwargs):
    super().refresh(**kwargs)
    
    self.from_binary = self.rng.choice([True, False])
    self.number_of_hexits = self.rng.randint(1, 8)
    self.value = self.rng.randint(1, 16**self.number_of_hexits)
    
    self.hex_val = f"0x{self.value:0{self.number_of_hexits}X}"
    self.binary_val = f"0b{self.value:0{4*self.number_of_hexits}b}"
    
    if self.from_binary:
      self.answers['answer'] = AnswerTypes.String(self.hex_val,
                                             label="Value in hex")
    else:
      self.answers['answer'] = AnswerTypes.String(self.binary_val,
                                             label="Value in binary")
  
  def _get_body(self, **kwargs):
    """Build question body and collect answers."""
    answers = [self.answers['answer']]

    body = ContentAST.Section()

    body.add_element(
      ContentAST.Paragraph([
        f"Given the number {self.hex_val if not self.from_binary else self.binary_val} "
        f"please convert it to {'hex' if self.from_binary else 'binary'}.",
        "Please include base indicator all padding zeros as appropriate (e.g. 0x01 should be 0b00000001)",
      ])
    )

    body.add_element(ContentAST.AnswerBlock(self.answers['answer']))

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body

  def _get_explanation(self, **kwargs):
    explanation = ContentAST.Section()
    
    paragraph = ContentAST.Paragraph([
      "The core idea for converting between binary and hex is to divide and conquer.  "
      "Specifically, each hexit (hexadecimal digit) is equivalent to 4 bits.  "
    ])
    
    if self.from_binary:
      paragraph.add_line(
        "Therefore, we need to consider each group of 4 bits together and convert them to the appropriate hexit."
      )
    else:
      paragraph.add_line(
        "Therefore, we need to consider each hexit and convert it to the appropriate 4 bits."
      )
    
    explanation.add_element(paragraph)
    
    # Generate translation table
    binary_str = f"{self.value:0{4*self.number_of_hexits}b}"
    hex_str = f"{self.value:0{self.number_of_hexits}X}"
    
    explanation.add_element(
      ContentAST.Table(
        data=[
          ["0b"] + [binary_str[i:i+4] for i in range(0, len(binary_str), 4)],
          ["0x"] + list(hex_str)
        ],
        # alignments='center', #['center' for _ in range(0, 1+len(hex_str))],
        padding=False
        
      )
    )
    
    if self.from_binary:
      explanation.add_element(
        ContentAST.Paragraph([
        f"Which gives us our hex value of: 0x{hex_str}"
        ])
      )
    else:
      explanation.add_element(
        ContentAST.Paragraph([
          f"Which gives us our binary value of: 0b{binary_str}"
        ])
      )

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation


@QuestionRegistry.register()
class AverageMemoryAccessTime(MathQuestion):
  
  CHANCE_OF_99TH_PERCENTILE = 0.75
  
  def refresh(self, rng_seed=None, *args, **kwargs):
    super().refresh(rng_seed=rng_seed, *args, **kwargs)
    
    # Figure out how many orders of magnitude different we are
    orders_of_magnitude_different = self.rng.randint(1,4)
    self.hit_latency = self.rng.randint(1,9)
    self.miss_latency = int(self.rng.randint(1, 9) * math.pow(10, orders_of_magnitude_different))
    
    # Add in a complication of making it sometimes very, very close
    if self.rng.random() < self.CHANCE_OF_99TH_PERCENTILE:
      # Then let's make it very close to 99%
      self.hit_rate = (99 + self.rng.random()) / 100
    else:
      self.hit_rate = self.rng.random()
      
    # Calculate the hit rate
    self.hit_rate = round(self.hit_rate, 4)
    
    # Calculate the AverageMemoryAccessTime (which is the answer itself)
    self.amat = self.hit_rate * self.hit_latency + (1 - self.hit_rate) * self.miss_latency
    
    self.answers = {
      "amat": AnswerTypes.Float(self.amat, label="Average Memory Access Time", unit="cycles")
    }
    
    # Finally, do the self.rngizing of the question, to avoid these being non-deterministic
    self.show_miss_rate = self.rng.random() > 0.5
    
    # At this point, everything in the question should be set.
    pass
  
  def _get_body(self, **kwargs):
    """Build question body and collect answers."""
    answers = [self.answers["amat"]]

    body = ContentAST.Section()

    # Add in background information
    body.add_element(
      ContentAST.Paragraph([
        ContentAST.Text("Please calculate the Average Memory Access Time given the below information. "),
        ContentAST.Text(
          f"Please round your answer to {ContentAST.Answer.DEFAULT_ROUNDING_DIGITS} decimal points. ",
          hide_from_latex=True
        )
      ])
    )
    table_data = [
      ["Hit Latency", f"{self.hit_latency} cycles"],
      ["Miss Latency", f"{self.miss_latency} cycles"]
    ]

    # Add in either miss rate or hit rate -- we only need one of them
    if self.show_miss_rate:
      table_data.append(["Miss Rate", f"{100 * (1 - self.hit_rate): 0.2f}%"])
    else:
      table_data.append(["Hit Rate", f"{100 * self.hit_rate: 0.2f}%"])

    body.add_element(
      ContentAST.Table(
        data=table_data
      )
    )

    body.add_element(ContentAST.LineBreak())

    body.add_element(ContentAST.AnswerBlock(self.answers["amat"]))

    return body, answers

  def get_body(self, **kwargs) -> ContentAST.Section:
    """Build question body (backward compatible interface)."""
    body, _ = self._get_body(**kwargs)
    return body

  def _get_explanation(self, **kwargs):
    explanation = ContentAST.Section()
    
    # Add in General explanation
    explanation.add_element(
      ContentAST.Paragraph([
        "Remember that to calculate the Average Memory Access Time "
        "we weight both the hit and miss times by their relative likelihood.",
        "That is, we calculate:"
      ])
    )
    
    # Add in equations
    explanation.add_element(
      ContentAST.Equation.make_block_equation__multiline_equals(
        lhs="AMAT",
        rhs=[
          r"(hit\_rate)*(hit\_cost) + (1 - hit\_rate)*(miss\_cost)",
          f"({self.hit_rate: 0.{ContentAST.Answer.DEFAULT_ROUNDING_DIGITS}f})*({self.hit_latency}) + ({1 - self.hit_rate: 0.{ContentAST.Answer.DEFAULT_ROUNDING_DIGITS}f})*({self.miss_latency}) = {self.amat: 0.{ContentAST.Answer.DEFAULT_ROUNDING_DIGITS}f}\\text{{cycles}}"
        ]
      )
    )

    return explanation, []

  def get_explanation(self, **kwargs) -> ContentAST.Section:
    """Build question explanation (backward compatible interface)."""
    explanation, _ = self._get_explanation(**kwargs)
    return explanation
  


#!env python
import abc
import logging
import math
from typing import List

from QuizGenerator.question import Question, QuestionRegistry
from QuizGenerator.contentast import ContentAST, AnswerTypes
from QuizGenerator.mixins import MathOperationQuestion

log = logging.getLogger(__name__)


class VectorMathQuestion(MathOperationQuestion, Question):

  def __init__(self, *args, **kwargs):
    kwargs["topic"] = kwargs.get("topic", Question.Topic.MATH)
    super().__init__(*args, **kwargs)

  def _generate_vector(self, dimension, min_val=-10, max_val=10):
    """Generate a vector with random integer values."""
    return [self.rng.randint(min_val, max_val) for _ in range(dimension)]

  def _format_vector(self, vector):
    """Return a ContentAST.Matrix element for the vector (format-independent).

    The Matrix element will render appropriately for each output format:
    - HTML: LaTeX bmatrix (for MathJax)
    - Typst: mat() with square bracket delimiter
    - LaTeX: bmatrix environment
    """
    # Convert to column matrix format: [[v1], [v2], [v3]]
    matrix_data = [[v] for v in vector]
    return ContentAST.Matrix(data=matrix_data, bracket_type="b")

  def _format_vector_inline(self, vector):
    """Format vector for inline display."""
    elements = [str(v) for v in vector]
    return f"({', '.join(elements)})"

  # Implement MathOperationQuestion abstract methods

  def generate_operands(self):
    """Generate two vectors for the operation."""
    if not hasattr(self, 'dimension'):
      self.dimension = self.rng.randint(self.MIN_DIMENSION, self.MAX_DIMENSION)
    vector_a = self._generate_vector(self.dimension)
    vector_b = self._generate_vector(self.dimension)
    return vector_a, vector_b

  def format_operand_latex(self, operand):
    """Format a vector for LaTeX display."""
    return self._format_vector(operand)

  def format_single_equation(self, operand_a, operand_b):
    """Format the equation for single questions."""
    operand_a_latex = self.format_operand_latex(operand_a)
    operand_b_latex = self.format_operand_latex(operand_b)
    return f"{operand_a_latex} {self.get_operator()} {operand_b_latex}"

  # Vector-specific overrides

  def refresh(self, *args, **kwargs):
    # Generate vector dimension first
    self.dimension = self.rng.randint(self.MIN_DIMENSION, self.MAX_DIMENSION)

    # Call parent refresh which will use our generate_operands method
    super().refresh(*args, **kwargs)

    self.vector_a = self.operand_a
    self.vector_b = self.operand_b

  def generate_subquestion_data(self):
    """Generate LaTeX content for each subpart of the question.
    Override to handle vector-specific keys in subquestion_data."""
    subparts = []
    for data in self.subquestion_data:
      # Map generic operand names to vector names for compatibility
      vector_a = data.get('vector_a', data['operand_a'])
      vector_b = data.get('vector_b', data['operand_b'])

      vector_a_latex = self._format_vector(vector_a)
      vector_b_latex = self._format_vector(vector_b)
      # Return as tuple of (matrix_a, operator, matrix_b)
      subparts.append((vector_a_latex, self.get_operator(), vector_b_latex))
    return subparts

  # Abstract methods that subclasses must still implement
  @abc.abstractmethod
  def get_operator(self):
    """Return the LaTeX operator for this operation."""
    pass

  @abc.abstractmethod
  def calculate_single_result(self, vector_a, vector_b):
    """Calculate the result for a single question with two vectors."""
    pass

  @abc.abstractmethod
  def create_subquestion_answers(self, subpart_index, result):
    """Create answer objects for a subquestion result."""
    pass


@QuestionRegistry.register()
class VectorAddition(VectorMathQuestion):

  MIN_DIMENSION = 2
  MAX_DIMENSION = 4

  def get_operator(self):
    return "+"

  def calculate_single_result(self, vector_a, vector_b):
    return [vector_a[i] + vector_b[i] for i in range(len(vector_a))]

  def create_subquestion_answers(self, subpart_index, result):
    raise NotImplementedError("Multipart not supported")

  def create_single_answers(self, result):
    self.answers["result"] = AnswerTypes.Vector(result)

  def _get_body(self):
    """Build question body and collect answers."""
    body = ContentAST.Section()

    body.add_element(ContentAST.Paragraph([self.get_intro_text()]))

    # Equation display using MathExpression for format-independent rendering
    vector_a_elem = self._format_vector(self.vector_a)
    vector_b_elem = self._format_vector(self.vector_b)
    body.add_element(ContentAST.MathExpression([
        vector_a_elem,
        " + ",
        vector_b_elem,
        " = "
    ]))

    # Canvas-only answer field - use stored answer for consistent UUID
    answer = self.answers["result"]
    body.add_element(ContentAST.OnlyHtml([ContentAST.Paragraph(["Enter your answer as a column vector:"])]))
    body.add_element(ContentAST.OnlyHtml([answer]))

    return body, list(self.answers.values())

  def _get_explanation(self):
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph(["To add vectors, we add corresponding components:"]))

    # Use LaTeX syntax for make_block_equation__multiline_equals
    vector_a_str = r" \\ ".join([str(v) for v in self.vector_a])
    vector_b_str = r" \\ ".join([str(v) for v in self.vector_b])
    result_str = r" \\ ".join([str(v) for v in self.result])
    addition_str = r" \\ ".join([f"{self.vector_a[i]}+{self.vector_b[i]}" for i in range(self.dimension)])

    explanation.add_element(
        ContentAST.Equation.make_block_equation__multiline_equals(
            lhs=r"\vec{a} + \vec{b}",
            rhs=[
                f"\\begin{{bmatrix}} {vector_a_str} \\end{{bmatrix}} + \\begin{{bmatrix}} {vector_b_str} \\end{{bmatrix}}",
                f"\\begin{{bmatrix}} {addition_str} \\end{{bmatrix}}",
                f"\\begin{{bmatrix}} {result_str} \\end{{bmatrix}}"
            ]
        )
    )

    return explanation, []


@QuestionRegistry.register()
class VectorScalarMultiplication(VectorMathQuestion):

  MIN_DIMENSION = 2
  MAX_DIMENSION = 4

  def _generate_scalar(self):
    """Generate a non-zero scalar for multiplication."""
    scalar = self.rng.randint(-5, 5)
    while scalar == 0:
      scalar = self.rng.randint(-5, 5)
    return scalar

  def refresh(self, *args, **kwargs):
    # Generate scalar first, then call parent refresh
    self.scalar = self._generate_scalar()
    super().refresh(*args, **kwargs)

  def get_operator(self):
    return f"{self.scalar} \\cdot"

  def calculate_single_result(self, vector_a, vector_b):
    return [self.scalar * component for component in vector_a]

  def create_subquestion_answers(self, subpart_index, result):
    raise NotImplementedError("Multipart not supported")

  def create_single_answers(self, result):
    self.answers["result"] = AnswerTypes.Vector(result)

  def _get_body(self):
    """Build question body and collect answers."""
    body = ContentAST.Section()

    body.add_element(ContentAST.Paragraph([self.get_intro_text()]))

    # Equation display using MathExpression
    vector_elem = self._format_vector(self.vector_a)
    body.add_element(ContentAST.MathExpression([
        f"{self.scalar} \\cdot ",
        vector_elem,
        " = "
    ]))

    # Canvas-only answer field - use stored answer
    answer = self.answers["result"]
    body.add_element(ContentAST.OnlyHtml([ContentAST.Paragraph(["Enter your answer as a column vector:"])]))
    body.add_element(ContentAST.OnlyHtml([answer]))

    return body, list(self.answers.values())

  def _get_explanation(self):
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph(["To multiply a vector by a scalar, we multiply each component by the scalar:"]))

    vector_str = r" \\ ".join([str(v) for v in self.vector_a])
    multiplication_str = r" \\ ".join([f"{self.scalar} \\cdot {v}" for v in self.vector_a])
    result_str = r" \\ ".join([str(v) for v in self.result])

    explanation.add_element(
        ContentAST.Equation.make_block_equation__multiline_equals(
            lhs=f"{self.scalar} \\cdot \\vec{{v}}",
            rhs=[
                f"{self.scalar} \\cdot \\begin{{bmatrix}} {vector_str} \\end{{bmatrix}}",
                f"\\begin{{bmatrix}} {multiplication_str} \\end{{bmatrix}}",
                f"\\begin{{bmatrix}} {result_str} \\end{{bmatrix}}"
            ]
        )
    )

    return explanation, []


@QuestionRegistry.register()
class VectorDotProduct(VectorMathQuestion):

  MIN_DIMENSION = 2
  MAX_DIMENSION = 4

  def get_operator(self):
    return "\\cdot"

  def calculate_single_result(self, vector_a, vector_b):
    return sum(vector_a[i] * vector_b[i] for i in range(len(vector_a)))

  def create_subquestion_answers(self, subpart_index, result):
    raise NotImplementedError("Multipart not supported")

  def create_single_answers(self, result):
    self.answers["dot_product"] = AnswerTypes.Int(result)

  def _get_body(self):
    """Build question body and collect answers."""
    body = ContentAST.Section()

    body.add_element(ContentAST.Paragraph([self.get_intro_text()]))

    # Equation display using MathExpression
    vector_a_elem = self._format_vector(self.vector_a)
    vector_b_elem = self._format_vector(self.vector_b)
    body.add_element(ContentAST.MathExpression([
        vector_a_elem,
        " \\cdot ",
        vector_b_elem,
        " = "
    ]))

    # Canvas-only answer field - use stored answer
    answer = self.answers["dot_product"]
    body.add_element(ContentAST.OnlyHtml([answer]))

    return body, list(self.answers.values())

  def _get_explanation(self):
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph(["The dot product is calculated by multiplying corresponding components and summing the results:"]))

    vector_a_str = r" \\ ".join([str(v) for v in self.vector_a])
    vector_b_str = r" \\ ".join([str(v) for v in self.vector_b])
    products_str = " + ".join([f"({self.vector_a[i]} \\cdot {self.vector_b[i]})" for i in range(self.dimension)])
    calculation_str = " + ".join([str(self.vector_a[i] * self.vector_b[i]) for i in range(self.dimension)])

    explanation.add_element(
        ContentAST.Equation.make_block_equation__multiline_equals(
            lhs="\\vec{a} \\cdot \\vec{b}",
            rhs=[
                f"\\begin{{bmatrix}} {vector_a_str} \\end{{bmatrix}} \\cdot \\begin{{bmatrix}} {vector_b_str} \\end{{bmatrix}}",
                products_str,
                calculation_str,
                str(self.result)
            ]
          )
      )

    return explanation, []  # Explanations don't have answers


@QuestionRegistry.register()
class VectorMagnitude(VectorMathQuestion):

  MIN_DIMENSION = 2
  MAX_DIMENSION = 3

  def get_operator(self):
    return "||"

  def calculate_single_result(self, vector_a, vector_b):
    magnitude_squared = sum(component ** 2 for component in vector_a)
    return math.sqrt(magnitude_squared)

  def create_subquestion_answers(self, subpart_index, result):
    raise NotImplementedError("Multipart not supported")

  def create_single_answers(self, result):
    self.answers["magnitude"] = AnswerTypes.Float(result)

  def _get_body(self):
    """Build question body and collect answers."""
    body = ContentAST.Section()

    body.add_element(ContentAST.Paragraph([self.get_intro_text()]))

    # Equation display using MathExpression
    vector_elem = self._format_vector(self.vector_a)
    body.add_element(ContentAST.MathExpression([
        "||",
        vector_elem,
        "|| = "
    ]))

    # Canvas-only answer field - use stored answer
    answer = self.answers["magnitude"]
    body.add_element(ContentAST.OnlyHtml([answer]))

    return body, list(self.answers.values())

  def _get_explanation(self):
    """Build question explanation."""
    explanation = ContentAST.Section()

    explanation.add_element(ContentAST.Paragraph(["The magnitude of a vector is calculated using the formula:"]))
    explanation.add_element(ContentAST.Equation(
        r"||\vec{v}|| = \sqrt{v_1^2 + v_2^2 + \cdots + v_n^2}", inline=False
    ))

    # Use LaTeX syntax for make_block_equation__multiline_equals
    vector_str = r" \\ ".join([str(v) for v in self.vector_a])
    squares_str = " + ".join([f"{v}^2" for v in self.vector_a])
    calculation_str = " + ".join([str(v**2) for v in self.vector_a])
    sum_of_squares = sum(component ** 2 for component in self.vector_a)
    result_formatted = sorted(ContentAST.Answer.accepted_strings(self.result), key=lambda s: len(s))[0]

    explanation.add_element(
        ContentAST.Equation.make_block_equation__multiline_equals(
            lhs=r"||\vec{v}||",
            rhs=[
                f"\\left|\\left| \\begin{{bmatrix}} {vector_str} \\end{{bmatrix}} \\right|\\right|",
                f"\\sqrt{{{squares_str}}}",
                f"\\sqrt{{{calculation_str}}}",
                f"\\sqrt{{{sum_of_squares}}}",
                result_formatted
            ]
        )
    )

    return explanation, []
